import pytest

from .test_api2_tags import temporary_sr_tag
from .test_api2_groups import temporary_group


@pytest.mark.asyncio
async def test_get_proximity_null(client):
    response = await client.get("/proximity")
    assert response.status_code == 200
    assert response.json() == []


@pytest.mark.asyncio
async def test_get_proximity(client):
    async with temporary_sr_tag() as tag_id:
        response = await client.get(f'/proximity')
        assert response.status_code == 200
        ls = response.json()
        assert len(ls) == 1

        response = await client.get('/proximity?offset=1')
        assert response.status_code == 200
        ls = response.json()
        assert len(ls) == 0

        response = await client.get('/listeners?limit=0')
        assert response.status_code == 200
        ls = response.json()
        assert len(ls) == 0


@pytest.mark.asyncio
async def test_proximity_filter_type(client):
    async with temporary_sr_tag() as new_id:
        response = await client.get('/proximity?type=SmartRelay')
        assert response.status_code == 200
        assert response.json()[0]['tag_id'] == new_id

        response = await client.get('/proximity?type=iBeacon')
        print('Response', response.json())
        assert response.status_code == 200
        assert len(response.json()) == 0


@pytest.mark.asyncio
async def test_tags_filter_group(client):
    async with temporary_group() as group_id:
        async with temporary_sr_tag() as new_id:
            response = await client.patch(f'/tags/{new_id}', json={
                'group_id': group_id
            })
            assert response.status_code == 204

            response = await client.get(f'/proximity?group_id={group_id}')
            assert response.status_code == 200
            assert response.json()[0]['tag_id'] == new_id

            response = await client.get(f'/proximity?group_id={group_id+1}')
            assert response.status_code == 200
            assert len(response.json()) == 0


@pytest.mark.asyncio
async def test_get_proximity_xml(client):
    async with temporary_sr_tag() as tag_id:
        response = await client.get(f'/proximity.xml')
        assert response.status_code == 200
        assert 'Content-Type' in response.headers
        assert response.headers['Content-Type'] == 'text/xml'
        assert len(response.text) > 20



"""Adds settings table and hysteresis

Revision ID: 9581ffb09767
Revises: 44f605b1cfa0
Create Date: 2021-04-23 12:50:15.579652

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = '9581ffb09767'
down_revision = '44f605b1cfa0'
branch_labels = None
depends_on = None


def upgrade():
    op.execute("""CREATE TABLE settings (
        name varchar,
        value varchar,
        PRIMARY KEY(name));""")
    op.execute('ALTER TABLE history ADD COLUMN id serial')
    op.execute('ALTER TABLE history ADD PRIMARY KEY(id)')
    op.execute("DROP FUNCTION append_history CASCADE")
    op.execute("""CREATE FUNCTION append_history()
    RETURNS TRIGGER
    LANGUAGE plpgsql
AS
$$
DECLARE
    curr_hist          history%ROWTYPE;
    prev_hist          history%ROWTYPE;
    min_time           interval;
    min_time_returning interval;
    tag_returning      boolean := FALSE;
BEGIN
    -- RAISE NOTICE 'In append_history():';
    -- RAISE NOTICE 'New Zone = %', NEW.zone_id;
    -- Fetch the last complete history entry for this tag
    SELECT * INTO prev_hist FROM history WHERE tag_id = OLD.id AND end_ts IS NOT NULL ORDER BY end_ts DESC LIMIT 1;
    IF FOUND THEN
        -- RAISE NOTICE 'Previous Zone = %', prev_hist.zone_id;
        IF prev_hist.zone_id = NEW.zone_id THEN
            -- We are trying to return to the previous zone, apply extra hysteresis
            -- RAISE NOTICE 'Returning';
            tag_returning := TRUE;
        END IF;
    END IF;

    -- Fetch current open history entry
    SELECT * INTO curr_hist FROM history WHERE tag_id = OLD.id AND end_ts IS NULL;
    IF FOUND THEN
        DECLARE
            duration_curr interval;
        BEGIN
            SELECT CONCAT(value, 'sec')::interval INTO min_time from settings WHERE name = 'MIN_SEC_IN_LOCATION';
            SELECT CONCAT(value, 'sec')::interval
            INTO min_time_returning
            FROM settings
            WHERE name = 'MIN_SEC_IN_LOCATION_IF_RETURNING';
            SELECT (now() at time zone 'utc') - curr_hist.start_ts INTO duration_curr;
            -- RAISE NOTICE 'Current Zone = %, duration = %', OLD.zone_id, duration_curr;
            -- IF tag_returning THEN
            --     RAISE NOTICE '% < % = %', duration_curr, min_time_returning, duration_curr < min_time_returning;
            -- ELSE
            --     RAISE NOTICE '% < % = %', duration_curr, min_time, duration_curr < min_time;
            -- END IF;
            IF (tag_returning = TRUE AND duration_curr < min_time_returning) OR
               (tag_returning = FALSE AND duration_curr < min_time) THEN
                -- We didn't stay here long enough, extend the previous history entry and delete current
                -- RAISE NOTICE 'Hit';
                UPDATE history SET end_ts = NULL WHERE id = prev_hist.id;
                DELETE FROM history WHERE id = curr_hist.id;
                IF tag_returning = TRUE THEN
                    -- Just re-open the previous history entry if returning otherwise we rely on the
                    -- code below to close out the previous entry and add the new one
                    -- RAISE NOTICE 'Returning hit';
                    return NEW;
                END IF;
            END IF;
        END;
    END IF;

    --RAISE NOTICE 'Pass through';
    UPDATE history SET end_ts = (now() at time zone 'utc') WHERE tag_id = OLD.id AND end_ts IS NULL;
    INSERT INTO history (tag_id, zone_id, distance, zone_name)
    VALUES (NEW.id, NEW.zone_id, NEW.distance,
            (SELECT zones.name from zones where zones.id = NEW.zone_id));
    RETURN NEW;
END;
$$""")  # noqa: E501
    op.execute("""CREATE TRIGGER tags_update_zone
            AFTER UPDATE OF zone_id ON tags FOR EACH ROW
            WHEN (OLD.zone_id IS DISTINCT FROM NEW.zone_id)
            EXECUTE PROCEDURE append_history();""")
    op.execute("DELETE FROM history where start_ts IS NULL")


def downgrade():
    op.execute("DROP TABLE settings;")
    op.execute("DROP FUNCTION append_history CASCADE")
    op.execute("""CREATE FUNCTION append_history()
            RETURNS TRIGGER
            LANGUAGE plpgsql
        AS $$
        BEGIN
            UPDATE history SET end_ts = (now() at time zone 'utc') WHERE tag_id = OLD.id AND end_ts IS NULL;
            INSERT INTO history (tag_id, zone_id, distance, zone_name)
            VALUES (NEW.id, NEW.zone_id, NEW.distance,
                    (SELECT zones.name from zones where zones.id = NEW.zone_id));
            RETURN NEW;
        END;
        $$""")  # noqa: E501
    op.execute("""CREATE TRIGGER tags_update_zone
            AFTER UPDATE OF zone_id ON tags FOR EACH ROW
            WHEN (OLD.zone_id IS DISTINCT FROM NEW.zone_id)
            EXECUTE PROCEDURE append_history();""")
    op.execute("DELETE FROM history where start_ts IS NULL")
    op.execute("ALTER TABLE history DROP COLUMN id")



import struct

import lzo
import pytest

from c3loc.config import CONFIG
from c3loc.ingest.frame import Frame, InvalidFrame
from c3loc.ingest.protocol import PacketType


def frame_message(packet_type, data, compress=False):
    flags = (1 << 7) if compress else 0
    flags |= packet_type & 0b01111111
    data = lzo.compress(data, 6, False) if compress else data
    header = struct.pack("!IB", len(data), flags)
    return b''.join([header, data])


@pytest.fixture
def frame():
    return Frame()


def test_frame_too_big(frame):
    msg = frame_message(PacketType.RAW,
                        b'0' * (CONFIG['MAX_FRAME_SIZE'] + 1),
                        False)
    with pytest.raises(InvalidFrame):
        frame.add_bytes(msg)


def test_bad_compressed_data(frame):
    data = b'0' * 50
    header = struct.pack("!IB", len(data), 1 << 7)  # first bit signals lzo
    msg = b''.join([header, data])
    with pytest.raises(InvalidFrame):
        frame.add_bytes(msg)


def test_multipacket_msg(frame):
    msg = frame_message(PacketType.RAW, b'\x00'*10, False)
    frame.add_bytes(msg[:2])
    assert len(frame.get_messages()) == 0
    frame.add_bytes(msg[2:])
    assert len(frame.get_messages()) == 1


def test_overflow_of_lzo_decompress_buf(frame):
    data = b'0' * (CONFIG['MAX_FRAME_SIZE'] + 500)
    assert len(data) > CONFIG['MAX_FRAME_SIZE']
    msg = frame_message(PacketType.RAW,
                        data,
                        True)
    assert len(msg) < CONFIG['MAX_FRAME_SIZE']
    with pytest.raises(InvalidFrame):
        frame.add_bytes(msg)



from __future__ import annotations
from datetime import datetime
from typing import Optional, Union

from pydantic import BaseModel, Field, validator  # type: ignore

from .types import DbInteger, SensorType, SensorValueType


class Sensors(BaseModel):
    tag_id: DbInteger
    ts: datetime
    type: int
    value: SensorValueType

    @classmethod
    def from_row(cls, row):
        vals = dict(row)
        typ: SensorType = vals['type']
        if typ == SensorType.SensorDataTemp:
            return SensorDataTemp(**vals)
        if typ == SensorType.SensorDataHumid:
            return SensorDataHumid(**vals)
        if typ == SensorType.SensorDataWetness:
            return SensorDataWetness(**vals)
        return None


class SensorDataTemp(BaseModel):
    tag_id: DbInteger
    ts: datetime
    value: float
    sensor_type = Field(default='SensorDataTemp', const=True)
    repr: Optional[str]

    @validator('value', pre=True)
    def convert_val(cls, v):
        return int.from_bytes(v, 'little', signed=True) / 10.0

    @validator('repr', always=True)
    def gen_repr(cls, v: Optional[str], values) -> str:
        if not v:
            return f'{values["value"]} ℃'
        return v


class SensorDataHumid(BaseModel):
    tag_id: DbInteger
    ts: datetime
    value: int
    sensor_type = Field(default='SensorDataHumid', const=True)
    repr: Optional[str]

    @validator('value', pre=True)
    def convert_val(cls, v):
        return int.from_bytes(v, 'little', signed=False)

    @validator('repr', always=True)
    def gen_repr(cls, v: Optional[str], values) -> str:
        if not v:
            return f'{values["value"]}% RH'
        return v


class SensorDataWetness(BaseModel):
    tag_id: DbInteger
    ts: datetime
    value: int
    sensor_type = Field(default='SensorDataWetness', const=True)
    repr: Optional[str]

    @validator('value', pre=True)
    def convert_val(cls, v):
        return int.from_bytes(v, 'little', signed=False)

    @validator('repr', always=True)
    def gen_repr(cls, v: Optional[str], values) -> str:
        if not v:
            return f'{values["value"]} Wetness'
        return v


AnySensor = Union[SensorDataTemp, SensorDataHumid, SensorDataWetness]



import sqlalchemy as sa  # type: ignore
from sqlalchemy.dialects.postgresql import UUID, MACADDR  # type: ignore

from sqlalchemy.ext.asyncio import (  # type: ignore
    AsyncConnection, create_async_engine)
from .config import DB_URL

engine = create_async_engine(DB_URL, echo=False)
metadata = sa.MetaData()

zones = sa.Table(
    "zones",
    metadata,
    sa.Column('id', sa.Integer, primary_key=True),
    sa.Column('name', sa.String(length=255)),
    sa.Column('attrs', sa.JSON, default='{}'))

groups = sa.Table(
    'groups',
    metadata,
    sa.Column('id', sa.Integer, primary_key=True),
    sa.Column('name', sa.String(length=255)),
    sa.Column('attrs', sa.JSON, default='{}'))

alarms = sa.Table(
    'alarms',
    metadata,
    sa.Column('id', sa.Integer, primary_key=True),
    sa.Column('tag_id', sa.ForeignKey('tags.id')),
    sa.Column('start_ts', sa.DateTime,
              default=sa.func.now()),
    sa.Column('last_ts', sa.DateTime,
              default=sa.func.now()),
    sa.Column('ack_ts', sa.DateTime),
    sa.Column('acknowledged', sa.Boolean, default=False),
    sa.Column('priority', sa.Integer, default=0)
)

tags = sa.Table(
    'tags',
    metadata,
    sa.Column('id', sa.Integer, primary_key=True),
    sa.Column('name', sa.String(length=255)),
    sa.Column('mac', MACADDR),
    sa.Column('uuid', UUID(as_uuid=True)),
    sa.Column('major', sa.Integer),
    sa.Column('minor', sa.Integer),
    sa.Column('type', sa.Enum(
        'iBeacon', 'SmartRelay', 'LocationAnchor',
        'SecureLocationAnchor', 'SecureSmartRelay', 'AlertWet',
        'MacBeacon',
        name='tag_type')),
    sa.Column('zone_id', sa.ForeignKey('zones.id')),
    sa.Column('last_seen', sa.DateTime,
              default=sa.func.now()),
    sa.Column('group_id', sa.ForeignKey('groups.id')),
    sa.Column('battery_pct', sa.Integer),
    sa.Column('distance', sa.Numeric(4, 2)),
    sa.Column('bid', sa.BigInteger),
    sa.Column('last_clock', sa.Integer),
    sa.Column('alarm_active', sa.Boolean, default=False),
    sa.Column('zone_start', sa.DateTime,
              default=sa.func.now()),
    sa.Column('attrs', sa.JSON, default={})
)

settings = sa.Table(
    'settings',
    metadata,
    sa.Column('name', sa.String(length=255), nullable=False,
              primary_key=True),
    sa.Column('value', sa.String(length=255), nullable=False),
)

sensors = sa.Table(
    'sensors',
    metadata,
    sa.Column('id', sa.Integer, primary_key=True),
    sa.Column('tag_id', sa.ForeignKey('tags.id'),
              nullable=False),
    sa.Column('type', sa.Integer, nullable=False),
    sa.Column('value', sa.LargeBinary(length=20), nullable=False),
    sa.Column('ts', sa.DateTime,
              default=sa.func.now()),
    sa.Column('device_ts', sa.Integer, nullable=False)
)

log = sa.Table(
    'log',
    metadata,
    sa.Column('tag_id', sa.ForeignKey('tags.id'),
              nullable=False),
    sa.Column('zone_id', sa.ForeignKey('zones.id')),
    sa.Column('ts', sa.DateTime,
              default=sa.func.now()),
    sa.Column('distance', sa.Numeric(4, 2)),
    sa.Column('variance', sa.Float),
    sa.Column('listener_id', sa.ForeignKey('listeners.id')),
    sa.Column('data', sa.JSON, default='{}'),
    sa.Column('anchor_dist', sa.Numeric(4, 2)),
    sa.Column('anchor_ts_delta', sa.Integer),
    sa.Column('anchor_id', sa.ForeignKey('tags.id')),
    sa.Column('reason', sa.Enum('ENTRY', 'MOVE', 'STATUS', 'EXIT',
                                name='report_reason')),
)

listeners = sa.Table(
    'listeners',
    metadata,
    sa.Column('id', sa.String(length=255), primary_key=True),
    sa.Column('name', sa.String(length=255)),
    sa.Column('zone_id', sa.ForeignKey('zones.id')),
    sa.Column('last_seen', sa.DateTime,
              default=sa.func.now()),
    sa.Column('attrs', sa.JSON, default='{}'),
    sa.UniqueConstraint('name', name='listener_name_key')
)

history = sa.Table(
    'history',
    metadata,
    sa.Column('tag_id', sa.ForeignKey('tags.id')),
    sa.Column('zone_id', sa.ForeignKey('zones.id')),
    sa.Column('zone_name', sa.String(length=255)),
    sa.Column('end_ts', sa.DateTime),
    sa.Column('distance', sa.Numeric(4, 2)),
    sa.Column('start_ts', sa.DateTime),
    sa.Column('id', sa.Integer, primary_key=True)
)


async def get_db() -> AsyncConnection:
    conn: AsyncConnection = await engine.connect()
    try:
        yield conn
        await conn.commit()
    except Exception as e:
        await conn.rollback()
        raise e
    finally:
        await conn.close()



"""adds history schema

Revision ID: 168e8731e481
Revises: 78e80c3339ff
Create Date: 2020-12-18 15:50:07.278444

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = '168e8731e481'
down_revision = '78e80c3339ff'
branch_labels = None
depends_on = None


def upgrade():
    op.execute("""CREATE TABLE IF NOT EXISTS history (
    tag_id int,
    zone_id int,
    zone_name varchar,
    ts timestamp without time zone default (now() at time zone 'utc'),
    distance numeric(4,2),
    CONSTRAINT fk_tag FOREIGN KEY (tag_id)
    REFERENCES tags(id) ON DELETE CASCADE,
    CONSTRAINT fk_zone FOREIGN KEY (zone_id)
    REFERENCES zones(id) ON DELETE SET NULL);""")

    op.execute("""CREATE FUNCTION append_history()
    RETURNS TRIGGER
    LANGUAGE plpgsql
AS $$
BEGIN
    INSERT INTO history (tag_id, zone_id, ts, distance, zone_name)
    VALUES (OLD.id, OLD.zone_id, OLD.last_seen, OLD.distance,
            (SELECT zones.name from zones where zones.id = OLD.zone_id));
    RETURN NEW;
END;
$$""")
    op.execute("""CREATE TRIGGER tags_update_zone
    AFTER UPDATE OF zone_id ON tags FOR EACH ROW
    WHEN (OLD.zone_id IS DISTINCT FROM NEW.zone_id AND OLD.zone_id IS NOT NULL)
    EXECUTE PROCEDURE append_history();""")
    op.execute("ALTER TABLE log RENAME distance_cm TO distance;")
    op.execute("ALTER TABLE tags RENAME distance_cm TO distance")


def downgrade():
    op.execute("DROP FUNCTION append_history CASCADE;")
    op.execute("DROP TABLE IF EXISTS history;")
    op.execute("ALTER TABLE log RENAME distance TO distance_cm")
    op.execute("ALTER TABLE tags RENAME distance TO distance_cm")



import pytest
import sqlalchemy as sa

from c3loc.db_next import alarms
from .test_api2_tags import new_sr_tag, delete_tag


@pytest.fixture
def delete_alarm(db, delete_tag):
    async def _del(alarm_id: int) -> None:
        result = await db.execute(sa.select(alarms.c.tag_id).where(
            alarms.c.id == alarm_id))
        tag_id = result.scalar()
        await db.execute(
            alarms.delete().where(alarms.c.id == alarm_id))
        await db.commit()
        await delete_tag(tag_id)
    return _del


@pytest.fixture
def new_alarm(db, new_sr_tag):
    async def _add() -> int:
        tag_id = await new_sr_tag()
        result = await db.execute(alarms.insert().values({
            'tag_id': tag_id
        }))
        new_id = result.inserted_primary_key[0]
        await db.commit()
        return new_id
    return _add


@pytest.fixture
async def temporary_alarm(db, new_alarm, delete_alarm):
    alarm_id = await new_alarm()
    yield alarm_id
    await delete_alarm(alarm_id)


@pytest.mark.asyncio
async def test_get_alarms_null(client):
    response = await client.get("/alarms")
    assert response.status_code == 200
    assert response.json() == []


@pytest.mark.asyncio
async def test_patch_alarm(client, temporary_alarm):
    new_id = temporary_alarm
    response = await client.patch(f'/alarms/{new_id}',
                                  json={'acknowledged': True})
    assert response.status_code == 204


@pytest.mark.asyncio
async def test_bad_patch_alarm(client, temporary_alarm):
    new_id = temporary_alarm
    response = await client.patch(f"/alarms/{new_id}", json={})
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_get_alarm(db, client, temporary_alarm):
    alarm_id = temporary_alarm
    response = await client.get(f'/alarms/{alarm_id}')
    assert response.status_code == 200


@pytest.mark.asyncio
async def test_get_alarms(client, temporary_alarm):
    alarm_id = temporary_alarm
    response = await client.get('/alarms')
    assert response.status_code == 200
    als = response.json()
    assert len(als) == 1
    assert als[0]['id'] == alarm_id

    response = await client.get('/alarms?offset=1')
    assert response.status_code == 200
    als = response.json()
    assert len(als) == 0

    response = await client.get('/alarms?limit=0')
    assert response.status_code == 200
    als = response.json()
    assert len(als) == 0


@pytest.mark.asyncio
async def test_alarms_404(client):
    response = await client.get('/alarms/1000')
    assert response.status_code == 404
    response = await client.patch('/alarms/1000', json={'acknowledged': True})
    assert response.status_code == 404


@pytest.mark.asyncio
async def test_alarms_no_post(client):
    response = await client.post('/alarms', json={})
    assert response.status_code == 405



from typing import Callable, Iterable

import fastapi  # type: ignore
from fastapi import Depends, APIRouter, Request
import sqlalchemy as sa  # type: ignore
from sqlalchemy.ext.asyncio import AsyncConnection  # type: ignore

from ..db_next import groups, get_db
from .dependencies import make_paginator
from .schemas.groups import Group, GroupCreate, GroupPatch

router = APIRouter()


@router.get('/groups', response_model=Iterable[Group])
async def get_groups(db: AsyncConnection = Depends(get_db),
                     p: Callable = Depends(make_paginator)):
    return await db.execute(p(sa.select(groups)))


@router.post('/groups')
async def create_group(group: GroupCreate, request: Request,
                       db: AsyncConnection = Depends(get_db)):
    insert = groups.insert().values(group.dict())
    result = await db.execute(insert)
    new_id = result.inserted_primary_key[0]
    root_path = request.scope.get("root_path")
    return fastapi.Response(status_code=201, headers={
        'Location': f'{root_path}/groups/{new_id}'})


@router.get('/groups/{group_id}', response_model=Group)
async def get_group(group_id: int, db: AsyncConnection = Depends(get_db)):
    result = await db.execute(groups.select().where(groups.c.id == group_id))
    entry = result.first()
    if not entry:
        raise fastapi.HTTPException(404)
    return entry


@router.delete('/groups/{group_id}')
async def delete_group(group_id: int, db: AsyncConnection = Depends(get_db)):
    result = await db.execute(groups.delete().where(groups.c.id == group_id))
    if result.rowcount == 0:
        raise fastapi.HTTPException(404)
    return fastapi.Response(status_code=204)


@router.patch('/groups/{group_id}')
async def patch_group(group_id: int, group_updates: GroupPatch,
                      db: AsyncConnection = Depends(get_db)):
    result = await db.execute(groups.update().values(
        group_updates.dict()).where(groups.c.id == group_id))
    if result.rowcount == 0:
        raise fastapi.HTTPException(404)
    return fastapi.Response(status_code=204)



import contextlib

from httpx import AsyncClient
import pytest
import sqlalchemy as sa

from c3loc.api2.main import app


@pytest.fixture
def delete_group(client):
    async def _del(group_id: int) -> None:
        await client.delete(f'/groups/{group_id}')
    return _del


@pytest.fixture
def new_group(client):
    async def _add(name: str = 'Test Group') -> int:
        group = {'name': name}
        response = await client.post('/groups', json=group)
        assert 'Location' in response.headers
        new_id = int(response.headers['Location'].split('/')[-1])
        return new_id
    return _add


@contextlib.asynccontextmanager
async def temporary_group(name: str = 'Test'):
    async with AsyncClient(app=app, base_url='http://test') as client:
        response = await client.post('/groups', json={
            'name': name})
    assert response.status_code == 201
    group_id = int(response.headers['Location'].split('/')[-1])
    print("Created Group: ", group_id)
    try:
        yield group_id
    finally:
        print("Destroying Group: ", group_id)
        async with AsyncClient(app=app, base_url='http://test') as client:
            await client.delete(f'/groups/{group_id}')


# @pytest.fixture
# async def temporary_group(new_group, delete_group):
#     group_id = await new_group()
#     yield group_id
#     await delete_group(group_id)


@pytest.mark.asyncio
async def test_get_groups_null(client):
    response = await client.get("/groups")
    assert response.status_code == 200
    assert response.json() == []


@pytest.mark.asyncio
async def test_post_group(client, db):
    group = {'name': 'Farts'}
    response = await client.post("/groups", json=group)
    assert response.status_code == 201
    await db.execute(sa.text('delete from groups'))


@pytest.mark.asyncio
async def test_patch_group(client):
    async with temporary_group() as new_id:
        response = await client.patch(f'/groups/{new_id}', json={'name': 'farts'})
    assert response.status_code == 204


@pytest.mark.asyncio
async def test_delete_group(client, new_group):
    new_id = await new_group()
    response = await client.delete(f'/groups/{new_id}')
    assert response.status_code == 204


@pytest.mark.asyncio
async def test_bad_post_patch_group(client):
    response = await client.post("/groups", json={})
    assert response.status_code == 422
    async with temporary_group() as new_id:
        response = await client.patch(f"/groups/{new_id}", json={})
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_get_group(client):
    async with temporary_group() as group_id:
        response = await client.get(f'/groups/{group_id}')
    assert response.status_code == 200


@pytest.mark.asyncio
async def test_get_groups(client):
    async with temporary_group() as group_id:
        response = await client.get(f'/groups')
        assert response.status_code == 200
        groups = response.json()
        assert len(groups) == 1
        assert groups[0]['id'] == group_id

        response = await client.get('/groups?offset=1')
        assert response.status_code == 200
        als = response.json()
        assert len(als) == 0

        response = await client.get('/groups?limit=0')
        assert response.status_code == 200
        als = response.json()
        assert len(als) == 0


@pytest.mark.asyncio
async def test_groups_404(client):
    response = await client.get('/groups/1000')
    assert response.status_code == 404
    response = await client.patch('/groups/1000', json={'name': 'farts'})
    assert response.status_code == 404
    response = await client.delete('/groups/1000')
    assert response.status_code == 404



"""Adds sensors table

Revision ID: 6b0896f05118
Revises: 5eae6a0253f5
Create Date: 2021-05-03 15:29:54.558118

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = '6b0896f05118'
down_revision = '5eae6a0253f5'
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """CREATE TABLE sensors
(
    id     serial,
    tag_id int,
    type   int,
    value  bytea,
    ts     timestamp without time zone DEFAULT (now() at time zone 'utc'),
    device_ts integer,
    PRIMARY KEY (id),
    CONSTRAINT fk_tags
        FOREIGN KEY (tag_id)
            REFERENCES tags (id) ON DELETE CASCADE)""")


def downgrade():
    op.execute('DROP TABLE sensors')



from binascii import unhexlify
from enum import Enum
import struct

from attrs import define, field
from . import SensorType


SAMPLE = unhexlify("0dff5900035f3fbe42877ded2bc60302e5fe")


class Battery(Enum):
    CR2032 = 1


@define
class Mopeka:
    battery_volts: float
    tank_fill_mm: int
    tank_quality: int
    sensor_id: int
    temperature: int
    tank_height_mm: int = field(kw_only=True, default=0)
    battery_type: Battery = field(kw_only=True, default=Battery.CR2032)

    @classmethod
    def from_hex(cls, h):
        return cls.from_bytes(unhexlify(h))

    @staticmethod
    def tank_process(tank1, tank2, temp):
        """ Takes raw beacon value and calculates mm height and confidence"""
        raw = (tank2 * 256) + tank1
        raw = raw & 0x3fff
        tank_height = int(
            raw * (0.573045 + (-0.002822 * temp) + (-0.00000535 * temp**2)))
        tank_confidence = (tank2 >> 6) & 0x3
        return (tank_height, tank_confidence)

    @classmethod
    def from_bytes(cls, b: bytes):
        (batt, temp, tank1, tank2, sensor_id) = struct.unpack(
            "<BBBB3s", b[5:12])
        batt = (batt & 0x7f) / 32
        temp = temp & 0x7f
        (tank_fill_mm, tank_quality) = cls.tank_process(tank1, tank2, temp)
        sensor_id = int.from_bytes(sensor_id, "little")
        temp = temp - 40
        return cls(batt, tank_fill_mm, tank_quality, sensor_id, temp)

    def battery_pct(self):
        mvolts = int(self.battery_volts * 1000)
        batt_pct = 0
        if mvolts >= 3000:
            batt_pct = 100
        elif mvolts > 2900:
            batt_pct = 100 - ((3000 - mvolts) * 58) / 100
        elif mvolts > 2740:
            batt_pct = 42 - ((2900 - mvolts) * 24) / 160
        elif mvolts > 2440:
            batt_pct = 18 - ((2740 - mvolts) * 12) / 300
        elif mvolts > 2100:
            batt_pct = 6 - ((2440 - mvolts) * 6) / 340
        return int(batt_pct)

    def to_sensor(self):
        return [(SensorType.BatteryPercent,
                 self.battery_pct().to_bytes(1, 'little')),
                (SensorType.Temperature,
                 (self.temperature*10).to_bytes(2, 'little', signed=True)),
                (SensorType.TankHeight,
                 self.tank_fill_mm.to_bytes(2, 'little')),
                (SensorType.TankConfidence,
                 self.tank_quality.to_bytes(1, 'little'))]



#!/bin/env python
"""Reference Implementation of v1 C3 Security Beacon Auth Server."""
from aiohttp import web
import aiohttp_cors  # type: ignore
import asyncio
import os
import signal

import click

from ..db import init_db_pool
from .alarms import AlarmsView, AlarmView
from .groups import GroupsView, GroupView
from .listeners import ListenersView, ListenerView
from .proximity import ProximityView, proximity_task, ProximityXMLView
from .sensors import SensorsView
from .server_meta import ServerMetaView
from .tags import (TagsView, TagView, TagHistoryView, TagContactView,
                   TagSensorView)
from .zones import ZonesView, ZoneView

from ..demo_fe.main import app as demo_fe_app


def default_handler(loop, context):
    print(f"Unhandled exception: {context['message']}", context)
    loop.stop()


@click.command()
@click.option('--port', '-p', default=5000)
@click.option('--demo-fe', is_flag=True, default=True)
def main(port, demo_fe) -> None:
    loop = asyncio.get_event_loop()
    loop.set_exception_handler(default_handler)
    if signal is not None and os.name != 'nt':
        loop.add_signal_handler(signal.SIGINT, loop.stop)

    async def init_db_app(app):
        app['db_pool'] = await init_db_pool('api')
        asyncio.ensure_future(proximity_task(app['db_pool']))

    app = web.Application()
    app.on_startup.append(init_db_app)
    app.router.add_view(r'/api', ServerMetaView)
    app.router.add_view(r'/api/listeners', ListenersView)
    app.router.add_view(
        r'/api/listeners/{id:\w+}', ListenerView, name='listener')
    app.router.add_view(r'/api/zones', ZonesView)
    app.router.add_view(r'/api/zones/{id:\d+}', ZoneView, name='zone')
    app.router.add_view(r'/api/tags', TagsView)
    app.router.add_view(r'/api/tags/{id:\d+}', TagView, name='tag')
    app.router.add_view(
        r'/api/tags/{id:\d+}/history', TagHistoryView, name='history')
    app.router.add_view(
        r'/api/tags/{id:\d+}/contacts', TagContactView, name='contacts')
    app.router.add_view(
        r'/api/tags/{id:\d+}/sensors', TagSensorView, name='sensors')
    app.router.add_view(r'/api/proximity', ProximityView)
    app.router.add_view(r'/api/proximity.xml', ProximityXMLView)
    app.router.add_view(r'/api/groups', GroupsView)
    app.router.add_view(r'/api/groups/{id:\d+}', GroupView, name='group')
    app.router.add_view(r'/api/alarms', AlarmsView)
    app.router.add_view(r'/api/alarms/{id:\d+}', AlarmView, name='alarm')
    app.router.add_view(r'/api/sensors', SensorsView)

    cors = aiohttp_cors.setup(app, defaults={
        "*": aiohttp_cors.ResourceOptions(
            allow_credentials=True,
            expose_headers="*",
            allow_headers="*",
        )
    })

    for route in list(app.router.routes()):
        cors.add(route)

    if demo_fe:
        app.add_subapp('/demo', demo_fe_app)
    web.run_app(app, port=port)


if __name__ == '__main__':
    main()



from typing import Callable, Iterable

import fastapi  # type: ignore
from fastapi import Depends, APIRouter  # type: ignore
import sqlalchemy as sa  # type: ignore
from sqlalchemy.ext.asyncio import AsyncConnection  # type: ignore

from ..db_next import alarms, get_db
import c3loc.api2.queries as q
from .dependencies import make_paginator
from .schemas.alarms import Alarm, AlarmPatch

router = APIRouter()


@router.get('/alarms', response_model=Iterable[Alarm])
async def get_alarms(db: AsyncConnection = Depends(get_db),
                     p: Callable = Depends(make_paginator)):
    query = q.get_alarms()
    result = await db.execute(p(query))
    return result


@router.get('/alarms/{alarm_id}', response_model=Alarm)
async def get_alarm(alarm_id: int, db: AsyncConnection = Depends(get_db)):
    result = await db.execute(q.get_alarm(alarm_id))
    entry = result.first()
    if not entry:
        raise fastapi.HTTPException(404)
    return entry


@router.patch('/alarms/{alarm_id}')
async def patch_alarm(alarm_id: int, alarm_updates: AlarmPatch,
                      db: AsyncConnection = Depends(get_db)):
    result = await db.execute(alarms.select().where(alarms.c.id == alarm_id))
    entry = result.first()
    if not entry:
        raise fastapi.HTTPException(404)
    if alarm_updates.acknowledged and not entry['ack_ts']:
        await db.execute(alarms.update().values(
            ack_ts=sa.func.now()).where(alarms.c.id == alarm_id))
    return fastapi.Response(status_code=204)



import pytest
import sqlalchemy as sa

from c3loc.db_next import listeners


@pytest.fixture
async def temporary_listener(engine):
    async with engine.connect() as conn:
        r = await conn.execute(
            listeners.insert().values({'id': 'aabbccddeeff',
                                       'attrs': {},
                                       'name': 'farts'}))
        await conn.commit()
        listener_id = r.inserted_primary_key[0]
    yield listener_id
    async with engine.connect() as conn:
        await conn.execute(
            listeners.delete().where(listeners.c.id == listener_id))
        await conn.commit()


@pytest.mark.asyncio
async def test_get_listeners_null(client):
    response = await client.get("/listeners")
    assert response.status_code == 200
    assert response.json() == []


@pytest.mark.asyncio
async def test_get_listener_default_name(client, engine):
    async with engine.connect() as conn:
        r = await conn.execute(
            listeners.insert().values({'id': 'aabbccddeeff',
                                       'attrs': {}}))
        await conn.commit()
        new_id = r.inserted_primary_key[0]

        resp = await client.get(f'/listeners/{new_id}')
        assert resp.status_code == 200
        assert resp.json()['name']


@pytest.mark.asyncio
async def test_patch_listener(client, temporary_listener):
    new_id = temporary_listener
    response = await client.patch(f'/listeners/{new_id}', json={'name': 'farts'})
    assert response.status_code == 204


@pytest.mark.asyncio
async def test_bad_patch_listener(client, temporary_listener):
    new_id = temporary_listener
    response = await client.patch(f"/listeners/{new_id}", json={'attrs': 1})
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_get_listener(client, temporary_listener):
    listener_id = temporary_listener
    response = await client.get(f'/listeners/{listener_id}')
    assert response.status_code == 200


@pytest.mark.asyncio
async def test_get_listeners(client, temporary_listener):
    listener_id = temporary_listener
    response = await client.get(f'/listeners')
    assert response.status_code == 200
    ls = response.json()
    assert len(ls) == 1
    assert ls[0]['id'] == listener_id

    response = await client.get('/listeners?offset=1')
    assert response.status_code == 200
    ls = response.json()
    assert len(ls) == 0

    response = await client.get('/listeners?limit=0')
    assert response.status_code == 200
    ls = response.json()
    assert len(ls) == 0


@pytest.mark.asyncio
async def test_listeners_404(client):
    response = await client.get('/listeners/1000')
    assert response.status_code == 404
    response = await client.patch('/listeners/1000', json={'name':'farts'})
    assert response.status_code == 404



from datetime import datetime
from typing import Dict, Optional

from pydantic import BaseModel, root_validator  # type: ignore

from .types import DbInteger, Attributes


class ListenerBase(BaseModel):
    zone_id: Optional[DbInteger]
    attrs: Optional[Attributes] = {}


class Listener(ListenerBase):
    id: str
    last_seen: datetime
    name: Optional[str]

    @root_validator
    def default_name(cls, values: Dict) -> Dict:
        if values.get('name', None):
            return values
        values['name'] = f'Listener {values["id"]}'
        return values


class ListenerPatch(ListenerBase):
    pass



"""Adds zone_start to tags

Revision ID: 5eae6a0253f5
Revises: 9581ffb09767
Create Date: 2021-04-29 10:58:04.477074

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = '5eae6a0253f5'
down_revision = '9581ffb09767'
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        "ALTER TABLE tags ADD COLUMN zone_start timestamp without time zone "
        "DEFAULT (now() at time zone 'UTC')")
    op.execute("DROP FUNCTION append_history CASCADE")
    op.execute("""CREATE FUNCTION append_history()
        RETURNS TRIGGER
        LANGUAGE plpgsql
    AS
    $$
    DECLARE
        curr_hist          history%ROWTYPE;
        prev_hist          history%ROWTYPE;
        min_time           interval;
        min_time_returning interval;
        tag_returning      boolean := FALSE;
    BEGIN
        -- RAISE NOTICE 'In append_history():';
        -- RAISE NOTICE 'New Zone = %', NEW.zone_id;
        -- Fetch the last complete history entry for this tag
        SELECT * INTO prev_hist FROM history WHERE tag_id = OLD.id AND end_ts IS NOT NULL ORDER BY end_ts DESC LIMIT 1;
        IF FOUND THEN
            -- RAISE NOTICE 'Previous Zone = %', prev_hist.zone_id;
            IF prev_hist.zone_id = NEW.zone_id THEN
                -- We are trying to return to the previous zone, apply extra hysteresis
                -- RAISE NOTICE 'Returning';
                tag_returning := TRUE;
            END IF;
        END IF;

        -- Fetch current open history entry
        SELECT * INTO curr_hist FROM history WHERE tag_id = OLD.id AND end_ts IS NULL;
        IF FOUND THEN
            DECLARE
                duration_curr interval;
            BEGIN
                SELECT CONCAT(value, 'sec')::interval INTO min_time from settings WHERE name = 'MIN_SEC_IN_LOCATION';
                SELECT CONCAT(value, 'sec')::interval
                INTO min_time_returning
                FROM settings
                WHERE name = 'MIN_SEC_IN_LOCATION_IF_RETURNING';
                SELECT (now() at time zone 'utc') - curr_hist.start_ts INTO duration_curr;
                -- RAISE NOTICE 'Current Zone = %, duration = %', OLD.zone_id, duration_curr;
                -- IF tag_returning THEN
                --     RAISE NOTICE '% < % = %', duration_curr, min_time_returning, duration_curr < min_time_returning;
                -- ELSE
                --     RAISE NOTICE '% < % = %', duration_curr, min_time, duration_curr < min_time;
                -- END IF;
                IF (tag_returning = TRUE AND duration_curr < min_time_returning) OR
                   (tag_returning = FALSE AND duration_curr < min_time) THEN
                    -- We didn't stay here long enough, extend the previous history entry and delete current
                    -- RAISE NOTICE 'Hit';
                    UPDATE history SET end_ts = NULL WHERE id = prev_hist.id;
                    DELETE FROM history WHERE id = curr_hist.id;
                    IF tag_returning = TRUE THEN
                        -- Just re-open the previous history entry if returning otherwise we rely on the
                        -- code below to close out the previous entry and add the new one
                        -- RAISE NOTICE 'Returning hit';
                        UPDATE tags SET zone_start = prev_hist.start_ts WHERE id = NEW.id;
                        return NEW;
                    END IF;
                END IF;
            END;
        END IF;

        --RAISE NOTICE 'Pass through';
        UPDATE history SET end_ts = (now() at time zone 'utc') WHERE tag_id = OLD.id AND end_ts IS NULL;
        INSERT INTO history (tag_id, zone_id, distance, zone_name)
        VALUES (NEW.id, NEW.zone_id, NEW.distance,
                (SELECT zones.name from zones where zones.id = NEW.zone_id));
        UPDATE tags SET zone_start = (now() at time zone 'utc') WHERE id = NEW.id;
        RETURN NEW;
    END;
    $$""")  # noqa: E501
    op.execute("""CREATE TRIGGER tags_update_zone
                AFTER UPDATE OF zone_id ON tags FOR EACH ROW
                WHEN (OLD.zone_id IS DISTINCT FROM NEW.zone_id)
                EXECUTE PROCEDURE append_history();""")


def downgrade():
    op.execute("ALTER TABLE tags DROP COLUMN zone_start")
    op.execute("DROP FUNCTION append_history CASCADE")
    op.execute("""CREATE FUNCTION append_history()
        RETURNS TRIGGER
        LANGUAGE plpgsql
    AS
    $$
    DECLARE
        curr_hist          history%ROWTYPE;
        prev_hist          history%ROWTYPE;
        min_time           interval;
        min_time_returning interval;
        tag_returning      boolean := FALSE;
    BEGIN
        -- RAISE NOTICE 'In append_history():';
        -- RAISE NOTICE 'New Zone = %', NEW.zone_id;
        -- Fetch the last complete history entry for this tag
        SELECT * INTO prev_hist FROM history WHERE tag_id = OLD.id AND end_ts IS NOT NULL ORDER BY end_ts DESC LIMIT 1;
        IF FOUND THEN
            -- RAISE NOTICE 'Previous Zone = %', prev_hist.zone_id;
            IF prev_hist.zone_id = NEW.zone_id THEN
                -- We are trying to return to the previous zone, apply extra hysteresis
                -- RAISE NOTICE 'Returning';
                tag_returning := TRUE;
            END IF;
        END IF;

        -- Fetch current open history entry
        SELECT * INTO curr_hist FROM history WHERE tag_id = OLD.id AND end_ts IS NULL;
        IF FOUND THEN
            DECLARE
                duration_curr interval;
            BEGIN
                SELECT CONCAT(value, 'sec')::interval INTO min_time from settings WHERE name = 'MIN_SEC_IN_LOCATION';
                SELECT CONCAT(value, 'sec')::interval
                INTO min_time_returning
                FROM settings
                WHERE name = 'MIN_SEC_IN_LOCATION_IF_RETURNING';
                SELECT (now() at time zone 'utc') - curr_hist.start_ts INTO duration_curr;
                -- RAISE NOTICE 'Current Zone = %, duration = %', OLD.zone_id, duration_curr;
                -- IF tag_returning THEN
                --     RAISE NOTICE '% < % = %', duration_curr, min_time_returning, duration_curr < min_time_returning;
                -- ELSE
                --     RAISE NOTICE '% < % = %', duration_curr, min_time, duration_curr < min_time;
                -- END IF;
                IF (tag_returning = TRUE AND duration_curr < min_time_returning) OR
                   (tag_returning = FALSE AND duration_curr < min_time) THEN
                    -- We didn't stay here long enough, extend the previous history entry and delete current
                    -- RAISE NOTICE 'Hit';
                    UPDATE history SET end_ts = NULL WHERE id = prev_hist.id;
                    DELETE FROM history WHERE id = curr_hist.id;
                    IF tag_returning = TRUE THEN
                        -- Just re-open the previous history entry if returning otherwise we rely on the
                        -- code below to close out the previous entry and add the new one
                        -- RAISE NOTICE 'Returning hit';
                        return NEW;
                    END IF;
                END IF;
            END;
        END IF;

        --RAISE NOTICE 'Pass through';
        UPDATE history SET end_ts = (now() at time zone 'utc') WHERE tag_id = OLD.id AND end_ts IS NULL;
        INSERT INTO history (tag_id, zone_id, distance, zone_name)
        VALUES (NEW.id, NEW.zone_id, NEW.distance,
                (SELECT zones.name from zones where zones.id = NEW.zone_id));
        RETURN NEW;
    END;
    $$""")  # noqa: E501
    op.execute("""CREATE TRIGGER tags_update_zone
                AFTER UPDATE OF zone_id ON tags FOR EACH ROW
                WHEN (OLD.zone_id IS DISTINCT FROM NEW.zone_id)
                EXECUTE PROCEDURE append_history();""")



import json

from aiohttp import web
from aiohttp_cors import CorsViewMixin  # type: ignore
from marshmallow import ValidationError

from ..config import CONFIG


class ValidatingView(web.View, CorsViewMixin):
    async def _valid_json(self, request):
        try:
            body = await self.request.json()
        except json.JSONDecodeError:
            raise web.HTTPBadRequest(reason='Invalid JSON')
        return body

    def _validate(self, schema, body, **kwargs):
        try:
            return schema.load(body, **kwargs)
        except ValidationError as e:
            item, problem = list(e.messages.items())[0]
            raise web.HTTPBadRequest(
                reason=f'Validation error on field "{item}": {problem[0]}')
        except json.JSONDecodeError:
            raise web.HTTPBadRequest(reason='Invalid JSON')


def paginate_query(request, query_tup, max_limit: int = None):
    if not max_limit:
        max_limit = int(CONFIG['API_RESULT_LIMIT'])
    try:
        offset = int(
            request.query['offset']) if 'offset' in request.query else 0
        limit = None
        if 'limit' in request.query:
            limit = int(request.query['limit'])
        else:
            limit = max_limit
        if limit > max_limit:
            limit = max_limit
    except ValueError:
        raise web.HTTPBadRequest(text='Bad pagination values')
    orig_query, *orig_args = query_tup
    new_query = (
        f'{orig_query} LIMIT ${len(orig_args)+1} OFFSET ${len(orig_args)+2}',
        *orig_args, limit, offset)
    return new_query



"""Adds SSR/SLA

Revision ID: fa1582c9a701
Revises: 168e8731e481
Create Date: 2021-02-10 15:56:52.096102

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = 'fa1582c9a701'
down_revision = '168e8731e481'
branch_labels = None
depends_on = None


def upgrade():
    op.execute("ALTER TABLE tags ADD COLUMN bid bigint UNIQUE;")
    op.execute("ALTER TABLE tags ADD COLUMN last_clock bigint;")
    op.execute("""
CREATE OR REPLACE FUNCTION def_tag_name() RETURNS TRIGGER LANGUAGE plpgsql AS $$
BEGIN
    IF NEW.name IS NOT NULL OR NEW.name != ''
       THEN RETURN NEW;
    ELSIF NEW.type = 'iBeacon' THEN
        NEW.name := 'iBeacon ' || NEW.uuid || ':' || NEW.major || ':' || NEW.minor;
    ELSIF NEW.type = 'LocationAnchor' THEN
            NEW.name := 'Location Anchor ' || NEW.major || ':' || NEW.minor;
    ELSIF NEW.type = 'SmartRelay' THEN
            NEW.name := 'SmartRelay ' || NEW.mac;
    ELSIF New.type = 'SecureSmartRelay' THEN
            NEW.name := 'SecureSmartRelay ' || NEW.bid;
    ELSIF New.type = 'SecureLocationAnchor' THEN
            NEW.name := 'SecureLocationAnchor ' || NEW.bid;
    END IF;
    return NEW;
END
$$;""")  # noqa: E501
    op.execute("""
DO $$ BEGIN
    ALTER TYPE tag_type ADD VALUE 'SecureLocationAnchor';
    ALTER TYPE tag_type ADD VALUE 'SecureSmartRelay';
EXCEPTION
    WHEN duplicate_object THEN null;
END $$;""")


def downgrade():
    op.execute("ALTER TABLE tags DROP COLUMN bid;")
    op.execute("ALTER TABLE tags DROP COLUMN last_clock;")
    op.execute("""
CREATE OR REPLACE FUNCTION def_tag_name() RETURNS TRIGGER LANGUAGE plpgsql AS $$
BEGIN
    IF NEW.name IS NOT NULL OR NEW.name != ''
       THEN RETURN NEW;
    ELSIF NEW.type = 'iBeacon' THEN
        NEW.name := 'iBeacon ' || NEW.uuid || ':' || NEW.major || ':' || NEW.minor;
    ELSIF NEW.type = 'LocationAnchor' THEN
            NEW.name := 'Location Anchor ' || NEW.major || ':' || NEW.minor;
    ELSIF NEW.type = 'SmartRelay' THEN
            NEW.name := 'SmartRelay ' || NEW.mac;
    END IF;
    return NEW;
END
$$;""")  # noqa: E501



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: c3loc/proto/common.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='c3loc/proto/common.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x18\x63\x33loc/proto/common.proto\"\\\n\nListenerID\x12\x13\n\x0blistener_id\x18\x01 \x01(\x0c\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0f\n\x07version\x18\x03 \x01(\t\x12\x0b\n\x03seq\x18\x04 \x01(\r\x12\r\n\x05model\x18\x05 \x01(\t\"\x18\n\tIDRequest\x12\x0b\n\x03seq\x18\x01 \x01(\rb\x06proto3'
)




_LISTENERID = _descriptor.Descriptor(
  name='ListenerID',
  full_name='ListenerID',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='listener_id', full_name='ListenerID.listener_id', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='ListenerID.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='version', full_name='ListenerID.version', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='seq', full_name='ListenerID.seq', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model', full_name='ListenerID.model', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28,
  serialized_end=120,
)


_IDREQUEST = _descriptor.Descriptor(
  name='IDRequest',
  full_name='IDRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='seq', full_name='IDRequest.seq', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=122,
  serialized_end=146,
)

DESCRIPTOR.message_types_by_name['ListenerID'] = _LISTENERID
DESCRIPTOR.message_types_by_name['IDRequest'] = _IDREQUEST
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ListenerID = _reflection.GeneratedProtocolMessageType('ListenerID', (_message.Message,), {
  'DESCRIPTOR' : _LISTENERID,
  '__module__' : 'c3loc.proto.common_pb2'
  # @@protoc_insertion_point(class_scope:ListenerID)
  })
_sym_db.RegisterMessage(ListenerID)

IDRequest = _reflection.GeneratedProtocolMessageType('IDRequest', (_message.Message,), {
  'DESCRIPTOR' : _IDREQUEST,
  '__module__' : 'c3loc.proto.common_pb2'
  # @@protoc_insertion_point(class_scope:IDRequest)
  })
_sym_db.RegisterMessage(IDRequest)


# @@protoc_insertion_point(module_scope)



from datetime import datetime, timedelta
from typing import Optional

import sqlalchemy as sa  # type: ignore
from dateutil.parser import isoparse
from fastapi import HTTPException
from pytz import utc

from ..config import CONFIG
from ..db_next import tags
from .schemas.tags import TagType
from .schemas.types import DbInteger

MAX_LIMIT: int = int(CONFIG['API_RESULT_LIMIT'])
CONTACTS_MAX_LIMIT: int = int(CONFIG['API_RESULT_LIMIT_CONTACTS'])


def make_paginator(limit: int = MAX_LIMIT, offset: int = 0):
    if limit < 0 or limit > MAX_LIMIT:
        limit = MAX_LIMIT
    offset = 0 if offset < 0 else offset

    def f(query):
        if offset:
            query = query.offset(offset)
        return query.limit(limit)
    return f


def make_contacts_paginator(limit: int = CONTACTS_MAX_LIMIT,
                            offset: int = 0):
    if limit < 0 or limit > CONTACTS_MAX_LIMIT:
        limit = CONTACTS_MAX_LIMIT
    offset = 0 if offset < 0 else offset

    def f(query):
        return query.bindparams(limit=limit, offset=offset)

    return f


def tag_filters(type: Optional[TagType] = None,
                group_id: Optional[DbInteger] = None):
    def f(query):
        if type:
            query = query.where(tags.c.type == type)
        if group_id:
            query = query.where(tags.c.group_id == group_id)
        return query
    return f


def bindable_time_filter(start: Optional[str] = None,
                         end: Optional[str] = None):
    def parse(x: Optional[str], name):
        if not x:
            return None
        try:
            y: datetime = isoparse(x)
        except ValueError:
            raise HTTPException(
                status_code=422, detail=f'Bad iso8601 in \'{name}\' parameter')
        if y.tzinfo:
            y = y.astimezone(utc).replace(tzinfo=None)
        return y

    s = parse(start, 'start') or datetime.utcnow() - timedelta(days=7)
    e = parse(end, 'end') or datetime.max

    def binder(text_expr):
        return text_expr.bindparams(start=s, end=e)
    return binder


def time_filters(start: Optional[str] = None,
                 end: Optional[str] = None):
    def parse(x: Optional[str], name):
        if not x:
            return None
        try:
            y: datetime = isoparse(x)
        except ValueError:
            raise HTTPException(
                status_code=422, detail=f'Bad iso8601 in \'{name}\' parameter')
        if y.tzinfo:
            y = y.astimezone(utc).replace(tzinfo=None)
        return y

    s = parse(start, 'start')
    e = parse(end, 'end')

    def f(query, start_field: Optional[sa.Column] = None,
          end_field: Optional[sa.Column] = None):
        if s and start_field is not None:
            query = query.where(start_field <= s)
        if e and end_field is not None:
            query = query.where(end_field >= e)
        return query
    return f



from __future__ import annotations
import asyncio
import json
import struct
from uuid import UUID

import attr
from cryptography.hazmat.primitives import cmac
from cryptography.hazmat.primitives.ciphers import algorithms

from ..config import CONFIG
from .frame import Frame, InvalidFrame
from ..proto.common_pb2 import ListenerID
from ..proto.location_pb2 import (IBeaconSummary, TempBeaconSummary,
                                  LABeaconSummary, RawAdvert)
from ..proto.secbeacon_pb2 import SecureSmartRelay
from .sensors import SensorReport
from .sensors.mopeka import Mopeka
import c3loc.stats as stats


ALERTWET_PREFIX = b'\x00\x86\x00\x03\x96\x06\x15\x00\x00\x00\x00\x11\x00\x00'
ALERTWET_THRESHOLD = 2**16 / 2


class PacketType:
    """Enum emulation class."""

    ID = 0
    DATA = 1
    SECURE = 2
    RESPONSE = 3
    REQUEST = 4
    ID_REQUEST = 5
    HEALTH_CHECK = 6
    TEMPBEACON = 7
    LABEACON = 8
    SSR = 9
    RAW = 10


class ADSubType:
    SmartRelay = 0
    SecureSmartRelay = 2
    SensorData = 3


def valid_ssr(ssr):
    key = CONFIG['SSR_KEY']
    c = cmac.CMAC(algorithms.AES(key))
    c.update(ssr.payload)
    calculated = c.finalize()
    if calculated[:4] == ssr.mac:
        stats.increment('Valid MAC')
        return True
    # print(calculated, ssr.payload.hex(), ssr.mac.hex())
    stats.increment('Failed MAC')
    return False


def start_packet_tasks():
    async def packet_task(t_id):
        print(f'Started packet task #{t_id}')
        while True:
            qd_pkt = await ListenerProtocol.packet_queue.get()
            try:
                await qd_pkt.protocol.process_packet(qd_pkt.msg)
            except Exception as e:
                stats.increment('Packet Processing Failure')
                print(f"Exception during packet_task: {e}")
                continue
            stats.increment('Packets Processed')
    for i in range(CONFIG['MAX_DB_CONNECTIONS']):
        asyncio.ensure_future(packet_task(i))


class ListenerProtocol(object):
    callback = None
    db_pool = None
    max_depth = int(CONFIG['MAX_QUEUED_PACKETS'])
    packet_queue: asyncio.Queue[QueuedPacket] = \
        asyncio.Queue(max_depth)
    packet_queue_reg_flag = False

    def __init__(self, verbose=False):
        self._rx_buf = bytearray()
        self.transport = None
        self._frame = Frame()
        self.verbose = verbose
        if not self.db_pool:
            raise NotImplementedError(
                "ListenerProtocol.db_pool must be set before instantiation")
        self.router = {
            PacketType.DATA: self.process_ibeacon,
            PacketType.LABEACON: self.process_relay_beacon,
            PacketType.SSR: self.process_ssr,
            PacketType.RAW: self.process_raw,
        }
        self.ignored_types = {
            PacketType.HEALTH_CHECK, PacketType.SECURE, PacketType.TEMPBEACON}
        self._l_id = None
        self.listener_id = None
        self.peer_name = None
        if not self.packet_queue_reg_flag:
            stats.register_cb(
                'Current Packet Queue Depth', self.packet_queue.qsize)
            self.packet_queue_reg_flag = True

    def connection_made(self, transport) -> None:
        self.transport = transport
        self.peer_name = transport.get_extra_info('peername')[0]
        print("Connection from {}".format(self.peer_name))
        stats.increment('Listener Connections')

    def eof_received(self):
        pass

    def data_received(self, data: bytes) -> None:
        try:
            self._frame.add_bytes(data)
        except InvalidFrame as e:
            print(f'Error: {self.peer_name} sent an InvalidFrame: '
                  f'{e.args[0]}. Disconnecting')
            stats.increment(f'Invalid Frame ({e.args[0]})')
            self.transport.close()
            return
        for msg in self._frame.get_messages():
            stats.increment("Packet Received")
            qd = QueuedPacket(msg, self)
            if self.packet_queue.full():
                stats.increment('Packet Dropped (Queue Full)')
                return
            self.packet_queue.put_nowait(qd)

    def _print_reason(self, report):
        print("\t\tReason: {}".format(self._reason(report)))

    def _reason(self, report):
        reason = ""
        if report.reason in [
                IBeaconSummary.IBeaconReport.EventReason.ENTRY,
                TempBeaconSummary.TempBeaconReport.EventReason.ENTRY,
                LABeaconSummary.LABeaconReport.EventReason.ENTRY]:
            reason = "ENTRY"
        if report.reason in [
                IBeaconSummary.IBeaconReport.EventReason.MOVE,
                TempBeaconSummary.TempBeaconReport.EventReason.MOVE,
                LABeaconSummary.LABeaconReport.EventReason.MOVE]:
            reason = "MOVE"
        if report.reason in [
                IBeaconSummary.IBeaconReport.EventReason.STATUS,
                TempBeaconSummary.TempBeaconReport.EventReason.STATUS,
                LABeaconSummary.LABeaconReport.EventReason.STATUS]:
            reason = "STATUS"
        if report.reason in [
                IBeaconSummary.IBeaconReport.EventReason.EXIT,
                TempBeaconSummary.TempBeaconReport.EventReason.EXIT,
                LABeaconSummary.LABeaconReport.EventReason.EXIT]:
            reason = "EXIT"
        return reason

    async def _upsert_listener(self, conn):
        await conn.execute(
            'INSERT into listeners (id) VALUES ($1) '
            'ON CONFLICT ON CONSTRAINT listeners_pkey '
            'DO UPDATE SET last_seen = (now() at time zone \'utc\') '
            'WHERE listeners.last_seen < now() - interval \'5sec\'',
            self.listener_id)
        # Add a default zone for listeners
        zone_id = await self._select_zone_id(conn)
        if not zone_id:
            r = await conn.fetchrow(
                'INSERT INTO zones (name) VALUES ($1) RETURNING id',
                f'Near Listener {self.listener_id}')
            zone_id = r[0]
            await conn.execute(
                'UPDATE listeners SET zone_id = $1 WHERE id = $2',
                zone_id, self.listener_id)

    async def _select_zone_id(self, conn):
        listener = await conn.fetchrow(
            'SELECT zone_id FROM listeners WHERE id = $1', self.listener_id)
        if not listener:
            return None
        return listener[0]

    async def process_packet(self, msg):
        if msg.type == PacketType.ID:
            self.process_id(msg)
            stats.increment(f'Packets Received ({self.listener_id})')
            return

        stats.increment(f'Packets Received ({self.listener_id})')

        if not self.listener_id:
            print(f'Error: {self.peer_name} non-ID packet before ID. '
                  f'Dropping connection')
            self.transport.close()
            return

        if msg.type in self.ignored_types:
            stats.increment('Ignored Packets')
            return

        conn = await self.db_pool.acquire()
        await self._upsert_listener(conn)
        if msg.type not in self.router:
            stats.increment('Unknown Packets')
            print(f"Warning: No implementation for packet type: {msg.type}")
        else:
            await self.router[msg.type](msg, conn)
        await self.db_pool.release(conn)

    def process_id(self, msg):
        l_id = ListenerID()
        l_id.ParseFromString(msg.data)
        self._l_id = l_id.listener_id
        self.listener_id = self._l_id.hex()
        print("{} identified as {} (version {})".format(
            self.peer_name, self._l_id.hex(), l_id.version))

    async def process_sensor(self, mac_addr, payload, conn):
        ts = payload[:4]
        ts = int.from_bytes(payload[:4], 'little', signed=False)
        payload = payload[4:]
        sensor = SensorReport.parse(payload)
        existing_tag_id = await conn.fetchval(
            "SELECT id from tags WHERE mac = $1 "
            "ORDER BY last_seen DESC LIMIT 1", mac_addr)
        if not existing_tag_id:
            stats.increment("Orphaned sensor packets")
            return
        insert_sensor = await conn.prepare(
            "INSERT INTO sensors (tag_id, type, value, device_ts) "
            "VALUES ($1, $2, $3, $4)")
        for field in sensor.fields:
            await insert_sensor.fetch(
                existing_tag_id, field.tag, field.value, ts)

    async def process_mopeka(self, data, conn):
        mac_addr = bytes(reversed(data.mac)).hex()
        # Parse fields
        m = Mopeka.from_bytes(data.payload)
        # Generate tag, if needed
        listener_zone = await self._select_zone_id(conn)
        (db_id,) = await conn.fetchrow(
            'INSERT into tags (bid, type, mac, zone_id) '
            'VALUES ($1, \'MacBeacon\', $2, $3) '
            'ON CONFLICT ON CONSTRAINT tags_bid_key '
            'DO UPDATE SET last_seen = (now() at time zone \'utc\'), '
            'zone_id = $3 RETURNING id', m.sensor_id, mac_addr,
            listener_zone)

        # Insert fields into sensors table
        insert_sensor = await conn.prepare(
            'INSERT INTO sensors (tag_id, type, value) '
            'VALUES ($1, $2, $3)')
        for sensor in m.to_sensor():
            await insert_sensor.fetch(
                db_id, *sensor)
        stats.increment("Mopeka packets")

    async def process_raw(self, msg, conn):
        data = RawAdvert()
        data.ParseFromString(msg.data)

        if data.payload.find(b'\xff\x59\x00') != -1:
            await self.process_mopeka(data, conn)
            return

        needle = b'\xff\x8d\x09'
        ad_start = data.payload.find(needle)
        payload = data.payload[ad_start+len(needle):]
        subtype = payload[0]

        if subtype == ADSubType.SensorData:
            stats.increment("Sensor Packets")
            mac_addr = bytes(reversed(data.mac)).hex()
            await self.process_sensor(mac_addr, payload[1:], conn)

    async def process_alertwet(self, b, conn):
        stats.increment('AlertWet Reports')

        (b_id,) = struct.unpack(">H", b.uuid[-2:])
        sensor = b.major_num
        batt = b.minor_num
        uuid = UUID(bytes=b.uuid)
        alarm_active = sensor > ALERTWET_THRESHOLD
        raw = json.dumps({'raw': f"{uuid}:{sensor}:{batt}",
                          'wetness': f"{sensor}",
                          'battery': f"{batt}"})
        listener_zone = await self._select_zone_id(conn)
        row = await conn.fetchrow(
            'INSERT into tags '
            '(bid, type, battery_pct, alarm_active, '
            'attrs, last_seen, zone_id, distance) '
            'VALUES ($1, \'AlertWet\', $2, $3, '
            '$4, (now() at time zone \'utc\'), $5, $6) '
            'ON CONFLICT ON CONSTRAINT tags_bid_key DO '
            'UPDATE SET battery_pct = excluded.battery_pct, '
            'alarm_active = excluded.alarm_active, '
            'attrs = excluded.attrs, '
            'last_seen = excluded.last_seen, '
            'zone_id = excluded.zone_id, '
            'distance = excluded.distance '
            'RETURNING tags.id',
            b_id, batt, alarm_active, raw, listener_zone, b.distance_cm / 100)
        (t_id,) = row

        await conn.fetch(
            'INSERT into log (tag_id, zone_id, listener_id, '
            'distance, variance, reason)'
            'VALUES ($1, $2, $3, $4, $5, $6)',
            t_id, listener_zone, self.listener_id, b.distance_cm / 100,
            round(b.variance / 2000, 1), self._reason(b))

        # Record sensor separately
        await conn.fetch(
            "INSERT into sensors (tag_id, type, value) "
            "VALUES  ($1, $2, $3)",
            t_id, 0x4, sensor.to_bytes(2, "little"))

    async def process_ibeacon(self, msg, conn):
        data = IBeaconSummary()
        data.ParseFromString(msg.data)

        # Convert to upsert when last_seen field is added?
        tag_insert = await conn.prepare(
            'INSERT into tags (uuid, major, minor, type)'
            'VALUES ($1, $2, $3, $4) '
            'RETURNING id, last_seen, (now() at time zone \'utc\')')
        tag_select = await conn.prepare(
            'SELECT id, last_seen, (now() at time zone \'utc\') FROM tags '
            'WHERE uuid = $1 AND major = $2 AND minor = $3')
        ibeacon_log = await conn.prepare(
            'INSERT into log '
            '(tag_id, zone_id, listener_id, distance, variance, reason)'
            'VALUES ($1, $2, $3, $4, $5, $6)')
        listener_zone = await self._select_zone_id(conn)
        for b in data.reports:

            if b.uuid.startswith(ALERTWET_PREFIX):
                await self.process_alertwet(b, conn)
                continue

            stats.increment('iBeacon Reports')
            uuid = UUID(bytes=b.uuid)
            tag_type = ''
            if uuid == CONFIG['LA_UUID']:
                tag_type = 'LocationAnchor'
            else:
                tag_type = 'iBeacon'
            t = await tag_select.fetchrow(uuid, b.major_num, b.minor_num)
            if not t:
                t = await tag_insert.fetchrow(uuid, b.major_num, b.minor_num,
                                              tag_type)
            t_id, t_lastseen, db_now = t[0], t[1], t[2]
            diff = (db_now - t_lastseen).seconds
            if diff > CONFIG['LAST_SEEN_RESOLUTION_SECS']:
                await conn.fetch(
                    'UPDATE tags SET last_seen = $1 WHERE id = $2',
                    db_now, t_id)
            await ibeacon_log.fetch(
                t_id, listener_zone, self.listener_id, b.distance_cm / 100,
                round(b.variance / 2000, 1), self._reason(b))

    async def process_ssr(self, msg, conn):
        data = SecureSmartRelay()
        data.ParseFromString(msg.data)
        if not valid_ssr(data):
            print("Invalid SSR, MAC check failed")
            return

        if len(data.payload) < 12:
            print("Ignoring short (but valid MAC?) SSR packet")
            return

        stats.increment("SSR Reports")
        (rssi1m, b_id, ts, status) = struct.unpack("<bIIB", data.payload[2:12])
        batt = status >> 1
        button = bool(status & 0x1)
        mac_addr = bytes(reversed(data.mac_addr)).hex()

        row = await conn.fetchrow(
            'INSERT into tags '
            '(bid, type, battery_pct, alarm_active, last_clock, mac) '
            'VALUES ($1, \'SecureSmartRelay\', $2, $3, $4, $5) '
            'ON CONFLICT ON CONSTRAINT tags_bid_key DO '
            'UPDATE SET battery_pct = excluded.battery_pct, '
            'alarm_active = excluded.alarm_active, '
            'last_clock = excluded.last_clock, '
            'mac = excluded.mac '
            'WHERE tags.last_clock < excluded.last_clock RETURNING tags.id',
            b_id, batt, button, ts, mac_addr)

        if not row:
            # print(f"Skipping duplicate or replayed packet for SSR {b_id}")
            return
        (t_id,) = row

        # Service the alarm state
        if button:
            row = await conn.fetchrow(
                'SELECT id from alarms '
                'WHERE tag_id = $1 and acknowledged = FALSE', t_id)
            if not row:
                await conn.fetchrow(
                    'INSERT INTO alarms (tag_id) VALUES ($1)', t_id)
            else:
                row = await conn.fetchrow(
                    'UPDATE alarms SET last_ts = now() '
                    'WHERE id = $1 AND acknowledged = FALSE', row[0])

        (anchor_id, anchor_delta, anchor_dist) = (None, None, None)
        azone_id = a_db_id = None
        if len(data.payload) == 18 and data.payload[1] in (0x0, 0x1):  # SSRv0
            (anchor_id, anchor_delta, anchor_dist) = struct.unpack(
                "<IBB", data.payload[12:])
            anchor_dist = anchor_dist / 10
        elif len(data.payload) == 19 and data.payload[1] == 0x2:  # SSRv1
            (anchor_id, anchor_delta, anchor_rssi, anchor_cal) = struct.unpack(
                "<IBBB", data.payload[12:])
            anchor_dist = 10 ** ((anchor_cal - anchor_rssi) / (10 * 3.2))

        if anchor_id:
            # Anchor upsert
            (a_db_id, azone_id) = await conn.fetchrow(
                'INSERT into tags (bid, type) '
                'VALUES ($1, \'SecureLocationAnchor\') '
                'ON CONFLICT ON CONSTRAINT tags_bid_key '
                'DO UPDATE SET last_seen = (now() at time zone \'utc\') '
                'RETURNING id, zone_id',
                anchor_id)
            if not azone_id:
                # Create default zone for anchor if nonexistant
                (azone_id,) = await conn.fetchrow(
                    'INSERT INTO zones (name) VALUES ($1) RETURNING id',
                    f'Near SLA {b_id}')
                await conn.execute(
                    'UPDATE tags SET zone_id = $1 WHERE id = $2',
                    azone_id, a_db_id)

        await conn.fetch(
            'INSERT into log (tag_id, zone_id, listener_id, distance,'
            'variance, anchor_id, anchor_dist, anchor_ts_delta, reason)'
            'VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9)',
            t_id, azone_id, self.listener_id, data.distance_cm / 100,
            round(data.variance / 2000, 1), a_db_id, anchor_dist, anchor_delta,
            None)

    async def process_relay_beacon(self, msg, conn):
        data = LABeaconSummary()
        data.ParseFromString(msg.data)
        # Convert to upsert when last_seen field is added?
        tag_insert = await conn.prepare(
            'INSERT into tags (mac, type, battery_pct) '
            'VALUES ($1, \'SmartRelay\', $2) RETURNING '
            'id, last_seen, (now() at time zone \'utc\'), alarm_active')
        tag_select = await conn.prepare(
            'SELECT id, last_seen, (now() at time zone \'utc\'), alarm_active '
            'FROM tags WHERE mac = $1 AND type = \'SmartRelay\'')
        tag_update = await conn.prepare(
            'UPDATE tags SET last_seen = $1, battery_pct = $3 WHERE id = $2')
        anchor_insert = await conn.prepare(
            'INSERT into tags (uuid, major, minor, type) '
            'VALUES ($1, $2, $3, \'LocationAnchor\') RETURNING id, zone_id')
        anchor_select = await conn.prepare(
            'SELECT id, zone_id FROM tags '
            'WHERE uuid = $1 AND major = $2 AND minor = $3')
        labeacon_log = await conn.prepare(
            'INSERT into log (tag_id, zone_id, listener_id, distance, '
            'variance, anchor_id, anchor_dist, anchor_ts_delta, reason)'
            'VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9)')
        alarm_update = await conn.prepare(
            'UPDATE alarms SET last_ts = now() '
            'WHERE tag_id = $1 AND acknowledged = FALSE RETURNING id')
        listener_zone = await self._select_zone_id(conn)
        for b in data.reports:
            stats.increment("SR Reports")
            b_id = bytes(reversed(b.beacon_id)).hex()

            t = await tag_select.fetchrow(b_id)
            if not t:
                stats.increment('New SR')
                t = await tag_insert.fetchrow(b_id, b.battery_pct)
            t_id, t_lastseen, db_now, t_alarm = t[0], t[1], t[2], t[3]
            diff = (db_now - t_lastseen).seconds
            if diff > CONFIG['LAST_SEEN_RESOLUTION_SECS']:
                await tag_update.fetch(db_now, t_id, b.battery_pct)

            # Service alarm
            alarm_active = b.status == \
                           LABeaconSummary.LABeaconReport.ButtonStatus.PRESSED
            if alarm_active:
                alarm = await alarm_update.fetchrow(t['id'])
                if not alarm:
                    stats.increment('New Alarm')
                    await conn.fetch(
                        'INSERT INTO alarms (tag_id) VALUES ($1)', t['id'])
            if t_alarm != alarm_active:
                stats.increment('Alarm State Updated')
                await conn.fetch(
                    'UPDATE tags SET alarm_active = $1 WHERE id = $2',
                    alarm_active, t_id)

            if b.nearest == 0 and b.nearest_delta == 0 and b.nearest_dist == 0:
                # No current anchor
                stats.increment('Unanchored SR Packet')
                await labeacon_log.fetch(
                    t['id'], listener_zone, self.listener_id,
                    b.distance_cm / 100, round(b.variance / 2000, 1), None,
                    None, None, self._reason(b))
                continue
            # Reconstruct the iBeacon from the LA data
            major = b.nearest >> 16
            minor = b.nearest & 0xffff
            r = await anchor_select.fetchrow(CONFIG['LA_UUID'], major, minor)
            if not r:
                stats.increment('New LA')
                r = await anchor_insert.fetchrow(
                    CONFIG['LA_UUID'], major, minor)
            (a_id, azone_id) = r[0], r[1]
            if not azone_id:
                stats.increment('Autocreated Zone (LA)')
                s = await conn.fetchrow(
                    'INSERT INTO zones (name) VALUES ($1) RETURNING id',
                    f'Near LA {major}:{minor}')
                azone_id = s[0]
                await conn.execute(
                    'UPDATE tags SET zone_id = $1 WHERE id = $2',
                    azone_id, a_id)
            await labeacon_log.fetch(
                t['id'], azone_id, self.listener_id, b.distance_cm / 100,
                round(b.variance / 2000, 1), a_id, b.nearest_dist / 10,
                b.nearest_delta, self._reason(b))

    async def process_health_check(self, msg, conn):
        pass

    def error_received(self, exc: Exception) -> None:
        stats.increment("Connection Errors")
        print(f'Error received: {exc}')

    def connection_lost(self, exc: Exception) -> None:
        stats.decrement("Listener Connections")
        print("{} has disconnected".format(self.peer_name))


@attr.s
class QueuedPacket:
    msg: bytes = attr.ib()
    protocol: ListenerProtocol = attr.ib()



from c3loc.ingest.sensors import SensorReport


def test_field_count():
    assert len(SensorReport.parse(b'').fields) == 0
    assert len(SensorReport.parse(b'\x01\x02\x00\x00').fields) == 1
    assert len(SensorReport.parse(b'\x02\x01\x00\x01\x01\x00').fields) == 2


def test_sensor_field():
    batt = SensorReport.parse(b'\x00\x01\x00').fields[0]
    assert batt.tag == 0
    assert batt.length == 1
    assert batt.value == b'\x00'
    temp = SensorReport.parse(b'\x01\x02\x00\x00').fields[0]
    assert temp.tag == 1
    assert temp.length == 2
    assert temp.value == b'\x00\x00'



from typing import Iterable

from fastapi import APIRouter, Depends  # type: ignore
from sqlalchemy.ext.asyncio import AsyncConnection  # type: ignore

from ..db_next import get_db
import c3loc.api2.queries as q
from .dependencies import make_paginator
from .schemas.sensors import Sensors, AnySensor

router = APIRouter()


@router.get('/sensors', response_model=Iterable[AnySensor])
async def sensors(db: AsyncConnection = Depends(get_db),
                  p=Depends(make_paginator)):
    result = await db.execute(p(q.get_sensors()))
    return (i for i in (Sensors.from_row(r) for r in result) if i)



import pkg_resources

from aiohttp import web
from aiohttp_cors import CorsViewMixin  # type: ignore

from ..config import CONFIG


class ServerMetaView(web.View, CorsViewMixin):
    async def get(self):
        be_version = pkg_resources.require("c3loc")[0].version
        return web.json_response({
            "api_version": "1.4.0",
            "backend_version": be_version,
            "api_result_limit": CONFIG['API_RESULT_LIMIT']})



"""add macbeacon mopeka

Revision ID: 1a641c271945
Revises: 1100e9db1a0d
Create Date: 2023-02-16 19:32:28.088031

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = '1a641c271945'
down_revision = '1100e9db1a0d'
branch_labels = None
depends_on = None


def upgrade():
    op.execute("""
CREATE OR REPLACE FUNCTION def_tag_name() RETURNS TRIGGER LANGUAGE plpgsql AS $$
BEGIN
    IF NEW.name IS NOT NULL OR NEW.name != ''
       THEN RETURN NEW;
    ELSIF NEW.type = 'iBeacon' THEN
        NEW.name := 'iBeacon ' || NEW.uuid || ':' || NEW.major || ':' || NEW.minor;
    ELSIF NEW.type = 'LocationAnchor' THEN
            NEW.name := 'Location Anchor ' || NEW.major || ':' || NEW.minor;
    ELSIF NEW.type = 'SmartRelay' THEN
            NEW.name := 'SmartRelay ' || NEW.mac;
    ELSIF New.type = 'SecureSmartRelay' THEN
            NEW.name := 'SecureSmartRelay ' || NEW.bid;
    ELSIF New.type = 'SecureLocationAnchor' THEN
            NEW.name := 'SecureLocationAnchor ' || NEW.bid;
    ELSIF New.type = 'AlertWet' THEN
            NEW.name := 'AlertWet ' || NEW.bid;
    ELSIF New.Type = 'MacBeacon' THEN
            NEW.name := 'MacBeacon ' || NEW.bid;
    END IF;
    return NEW;
END
$$;""")  # noqa: E501
    op.execute("""
DO $$ BEGIN
    ALTER TYPE tag_type ADD VALUE 'MacBeacon';
EXCEPTION
    WHEN duplicate_object THEN null;
END $$;""")  # noqa: E501


def downgrade():
    op.execute("""
CREATE OR REPLACE FUNCTION def_tag_name() RETURNS TRIGGER LANGUAGE plpgsql AS $$
BEGIN
    IF NEW.name IS NOT NULL OR NEW.name != ''
       THEN RETURN NEW;
    ELSIF NEW.type = 'iBeacon' THEN
        NEW.name := 'iBeacon ' || NEW.uuid || ':' || NEW.major || ':' || NEW.minor;
    ELSIF NEW.type = 'LocationAnchor' THEN
            NEW.name := 'Location Anchor ' || NEW.major || ':' || NEW.minor;
    ELSIF NEW.type = 'SmartRelay' THEN
            NEW.name := 'SmartRelay ' || NEW.mac;
    ELSIF New.type = 'SecureSmartRelay' THEN
            NEW.name := 'SecureSmartRelay ' || NEW.bid;
    ELSIF New.type = 'SecureLocationAnchor' THEN
            NEW.name := 'SecureLocationAnchor ' || NEW.bid;
    ELSIF New.type = 'AlertWet' THEN
            NEW.name := 'AlertWet ' || NEW.bid;
    END IF;
    return NEW;
END
$$;""")  # noqa: E501



from aiohttp import web
from aiohttp_cors import CorsViewMixin  # type: ignore
import asyncpg  # type: ignore
import asyncio

import jinja2

from .schemas import prox_schema
from .views import paginate_query
from ..config import CONFIG

THE_QUERY = """
SELECT tags.distance,
       zones.name as zone_name,
       tags.name as tag_name,
       tags.type as tag_type,
       last_seen,
       tags.id as tag_id,
       zone_id,
       alarm_active
FROM tags
LEFT JOIN zones on zones.id = zone_id
ORDER BY tags.id DESC
"""

THE_TYPE_QUERY = """
SELECT tags.distance,
       zones.name as zone_name,
       tags.name as tag_name,
       tags.type as tag_type,
       last_seen,
       tags.id as tag_id,
       zone_id,
       alarm_active
FROM tags
LEFT JOIN zones on zones.id = zone_id
WHERE tags.type = $1
ORDER BY tags.id DESC
"""

THE_GROUP_QUERY = """
SELECT tags.distance,
       zones.name as zone_name,
       tags.name as tag_name,
       tags.type as tag_type,
       last_seen,
       tags.id as tag_id,
       zone_id,
       alarm_active
FROM tags
LEFT JOIN zones on zones.id = zone_id
WHERE tags.group_id = $1
ORDER BY tags.id DESC
"""


def process_prox_results(results):
    output = []
    for r in results:
        zone_id = r['zone_id']
        d_r = dict(r)
        d_r['links'] = {
            "zone": f"/api/zones/{zone_id}" if zone_id else None,
            "tag": f"/api/tags/{r['tag_id']}"
        }
        output.append(d_r)
    return output


class ProximityView(web.View, CorsViewMixin):
    async def get(self):
        if 'type' in self.request.query:
            t = self.request.query['type']
            query = (THE_TYPE_QUERY, t)
        elif 'group_id' in self.request.query:
            g = self.request.query['group_id']
            try:
                query = (THE_GROUP_QUERY, int(g))
            except ValueError:
                raise web.HTTPBadRequest(text='Invalid group id')
        else:
            query = (THE_QUERY,)
        query = paginate_query(self.request, query)
        async with self.request.app['db_pool'].acquire() as conn:
            try:
                results = await conn.fetch(*query)
            except asyncpg.exceptions.InvalidTextRepresentationError:
                raise web.HTTPBadRequest(text='Bad tag type in query')
            return web.json_response(
                prox_schema.dump(process_prox_results(results)))


PROXIMITY_TASK_QUERY = """
UPDATE tags
SET zone_id  = cf.zone_id,
    distance = cf.distance
FROM (SELECT DISTINCT ON (upper.tag_id) coalesce(anchor_dist, upper.distance) as distance,
                                        upper.tag_id,
                                        tags.zone_id
      from log upper
               LEFT JOIN tags on tags.id = upper.anchor_id
      WHERE ts > tags.last_seen - interval '10s'
      ORDER BY upper.tag_id, distance, upper.ts DESC) as cf
WHERE tags.id = cf.tag_id
"""  # noqa: E501


async def proximity_task(pool):
    period = CONFIG['LOCATION_UPDATE_MS'] / 1000
    count = 0

    while True:
        async with pool.acquire() as conn:
            await conn.execute(PROXIMITY_TASK_QUERY)
            count += 1
            if count % 60 == 0:  # Once per minute
                await conn.execute(
                    'DELETE FROM log '
                    'WHERE ts < current_timestamp - interval \'5m\'')
            await asyncio.sleep(period)


XML_TEMPLATE = jinja2.Template("""<?xml version="1.0" encoding="UTF-8"?>
<TagData>
  {%- for tag in tags %}
  <Tag id="{{ tag.id }}">
    <Name>{{ tag.name }}</Name>
    {%- if tag.type == "SmartRelay" %}
    <MacAddress>{{ tag.mac }}</MacAddress>
    {%- endif %}
    <BatteryPercent>{{ tag.battery_pct }}</BatteryPercent>
    {%- if tag.zone_id %}
    <Location>{{ tag.zone_name }}</Location>
    <LocationId>{{ tag.zone_id }}</LocationId>
    <LastSeen>{{ tag.last_seen }}</LastSeen>
    <Distance>{{ tag.distance }}</Distance>
    <Duration>{{ tag.duration }}</Duration>
    {%- else %}
    <Location>Unknown</Location>
    {%- endif %}
  </Tag>
  {%- endfor %}
</TagData>
""")

XML_QUERY = """
SELECT tags.distance,
       zones.name as zone_name,
       tags.name,
       tags.mac,
       tags.type,
       tags.battery_pct,
       last_seen AT TIME ZONE 'UTC' as last_seen,
       tags.id,
       zone_id,
       alarm_active,
       hist.duration as duration
FROM tags
LEFT JOIN zones ON zones.id = zone_id
LEFT JOIN (
       SELECT tag_id, (now() at time zone 'utc') - start_ts AS duration
       FROM history WHERE end_ts IS NULL) hist on hist.tag_id = tags.id
WHERE tags.type IN ('SmartRelay', 'SecureSmartRelay')
ORDER BY tags.id DESC
"""


class ProximityXMLView(web.View, CorsViewMixin):
    async def get(self):
        results = None
        async with self.request.app['db_pool'].acquire() as conn:
            results = await conn.fetch(XML_QUERY)
        return web.Response(
            text=XML_TEMPLATE.render(tags=results), content_type="text/xml",
            charset="utf-8")



from binascii import unhexlify
import os
from typing import Dict, Union
from uuid import UUID

CONFIG: Dict[str, Union[str, int, UUID, bytes]] = {
    'DB_NAME': os.getenv('DB_NAME', ''),
    'DB_USER': os.getenv('DB_USER', 'c3loc'),
    'DB_PASSWORD': os.getenv('DB_PASSWORD', 'c3letmein'),
    'DB_HOST': os.getenv('DB_HOST', ''),
    'LA_UUID': os.getenv('LA_UUID', UUID(bytes=b'c3wirelesslocanc')),
    'API_RESULT_LIMIT': int(os.getenv('API_RESULT_LIMIT', 100)),
    'API_RESULT_LIMIT_CONTACTS':
        int(os.getenv('API_RESULT_LIMIT_CONTACTS', 5000)),
    'LOCATION_UPDATE_MS': 1000,
    'MAX_FRAME_SIZE': int(os.getenv('MAX_FRAME_SIZE', 5 * 1024 * 1024)),
    'SSR_KEY': unhexlify(os.getenv('SSR_KEY', b'c3wirelesslocanc'.hex())),
    'STATS_INTERVAL': int(os.getenv('STATS_INTERVAL', 60)),
    'MAX_DB_CONNECTIONS': int(os.getenv('MAX_DB_CONNECTIONS', 4)),
    'MAX_QUEUED_PACKETS': int(os.getenv('MAX_QUEUED_PACKETS', 100)),
    'LAST_SEEN_RESOLUTION_SECS':
        int(os.getenv('LAST_SEEN_RESOLUTION_SECS', 5)),
    'MIN_SEC_IN_LOCATION': int(os.getenv('MIN_SEC_IN_LOCATION', 10)),
    'MIN_SEC_IN_LOCATION_IF_RETURNING':
        int(os.getenv('MIN_SEC_IN_LOCATION_IF_RETURNING', 30)),
}

DB_RELEVANT = {
    'ingest': {'MIN_SEC_IN_LOCATION': CONFIG['MIN_SEC_IN_LOCATION'],
               'MIN_SEC_IN_LOCATION_IF_RETURNING':
                   CONFIG['MIN_SEC_IN_LOCATION_IF_RETURNING']},
    'api': {}
}

DB_PATH = os.getenv(
    'DB_PATH',
    f'{CONFIG["DB_USER"]}:{CONFIG["DB_PASSWORD"]}'  # type: ignore
    f'@{CONFIG["DB_HOST"]}/{CONFIG["DB_NAME"]}')

DB_URL = os.getenv(
    'DB_URL',
    f'postgresql+asyncpg://{DB_PATH}')  # type: ignore



"""drop uk_mac constraint from tags table

Revision ID: dbe74359375f
Revises: 6b0896f05118
Create Date: 2021-05-29 06:16:17.197295

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = 'dbe74359375f'
down_revision = '6b0896f05118'
branch_labels = None
depends_on = None


def upgrade():
    op.execute("ALTER TABLE tags DROP CONSTRAINT uk_mac")


def downgrade():
    op.execute("ALTER TABLE tags ADD CONSTRAINT uk_mac UNIQUE (mac)")



from typing import Iterable

import fastapi  # type: ignore
from fastapi import Depends, APIRouter
import sqlalchemy as sa  # type: ignore
from sqlalchemy.ext.asyncio import AsyncConnection  # type: ignore

from ..db_next import listeners, get_db
from .dependencies import make_paginator
from .schemas.listeners import Listener, ListenerPatch

router = APIRouter()


@router.get('/listeners', response_model=Iterable[Listener])
async def get_listeners(db: AsyncConnection = Depends(get_db),
                        p=Depends(make_paginator)):
    return await db.execute(p(sa.select(listeners)))


@router.get('/listeners/{listener_id}', response_model=Listener)
async def get_listener(listener_id: str,
                       db: AsyncConnection = Depends(get_db)):
    result = await db.execute(
        listeners.select().where(listeners.c.id == listener_id))
    entry = result.first()
    if not entry:
        raise fastapi.HTTPException(404)
    return entry


@router.patch('/listeners/{listener_id}')
async def patch_listener(listener_id: str, listener_updates: ListenerPatch,
                         db: AsyncConnection = Depends(get_db)):
    result = await db.execute(listeners.update().values(
        listener_updates.dict()).where(listeners.c.id == listener_id))
    if result.rowcount == 0:
        raise fastapi.HTTPException(404)
    return fastapi.Response(status_code=204)



from functools import partial

import pytz
from aiohttp import web
from aiohttp_cors import CorsViewMixin  # type: ignore

from .views import paginate_query
from ..ingest.sensors import SensorType


# class SensorType(enum.IntEnum):
#    Battery = 0
#    Temperature = 1
#    Humidity = 2


API_SENSOR_NAME_MAP = {
    SensorType.BatteryPercent: "SensorDataBattery",
    SensorType.Temperature: "SensorDataTemp",
    SensorType.Humidity: "SensorDataHumid",
    SensorType.TankHeight: "SensorTankHeight",
    SensorType.TankConfidence: "SensorTankConfidence",
}


def fixed_point(divisor, data, signed=True):
    val = int.from_bytes(data, 'little', signed=signed)
    return val / divisor


def percent_formatter(data):
    return int.from_bytes(data, 'little', signed=False)


def format_sensor(tag, data):
    dispatch = {SensorType.BatteryPercent: percent_formatter,
                SensorType.Temperature: partial(fixed_point, 10),
                SensorType.Humidity: percent_formatter,
                SensorType.TankHeight: partial(fixed_point, 1),
                SensorType.TankConfidence: partial(fixed_point, 1)}
    if tag not in dispatch:
        return None
    return dispatch[tag](data)


def format_sensors(sensors):
    out = []
    for sensor in sensors:
        ts = sensor['ts']
        value = format_sensor(sensor['type'], sensor['value'])
        if value is None:
            continue
        out.append({
            "sensor_type": API_SENSOR_NAME_MAP[sensor['type']],
            "ts": ts.replace(tzinfo=pytz.utc).isoformat(),
            "value": value
        })
    return out


SENSORS_QUERY = """SELECT * from (
SELECT DISTINCT ON (tag_id, type) *
FROM sensors
ORDER BY tag_id, type, ts DESC) as filter
"""


class SensorsView(web.View, CorsViewMixin):
    async def get(self):
        async with self.request.app['db_pool'].acquire() as conn:
            query = (SENSORS_QUERY,)
            query = paginate_query(self.request, query)
            sensors = await conn.fetch(*query)
            out = []
            for sensor in sensors:
                ts = sensor['ts']
                value = format_sensor(sensor['type'], sensor['value'])
                if value is None:
                    continue
                out.append({
                    "tag_id": sensor['tag_id'],
                    "sensor_type": API_SENSOR_NAME_MAP[sensor['type']],
                    "ts": ts.replace(tzinfo=pytz.utc).isoformat(),
                    "value": value
                })
            return web.json_response(out)


__all__ = ['API_SENSOR_NAME_MAP', 'format_sensors', 'SensorType',
           'SensorsView']



import contextlib

from httpx import AsyncClient
import pytest
import sqlalchemy as sa

from c3loc.api2.main import app


@pytest.fixture
def delete_zone(client):
    async def _del(zone_id: int) -> None:
        await client.delete(f'/zones/{zone_id}')
    return _del


@pytest.fixture
def new_zone(client):
    async def _add(name: str = 'Test Zone') -> int:
        zone = {'name': name}
        response = await client.post('/zones', json=zone)
        assert 'Location' in response.headers
        new_id = int(response.headers['Location'].split('/')[-1])
        return new_id
    return _add


@contextlib.asynccontextmanager
async def temporary_zone(name: str = 'Test'):
    print("Creating Zone")
    async with AsyncClient(app=app, base_url='http://test') as client:
        response = await client.post('/zones', json={
            'name': name})
    assert response.status_code == 201
    zone_id = int(response.headers['Location'].split('/')[-1])
    try:
        yield zone_id
    finally:
        print("Destroying Zone")
        async with AsyncClient(app=app, base_url='http://test') as client:
            await client.delete(f'/groups/{zone_id}')


# @pytest.fixture
# async def temporary_zone(client, new_zone, delete_zone):
#     zone_id = await new_zone()
#     yield zone_id
#     await delete_zone(zone_id)


@pytest.mark.asyncio
async def test_get_zones_null(client):
    response = await client.get("/zones")
    assert response.status_code == 200
    assert response.json() == []


@pytest.mark.asyncio
async def test_post_zone(client, db):
    zone = {'name': 'Test'}
    response = await client.post("/zones", json=zone)
    assert response.status_code == 201
    await db.execute(sa.text('delete from zones'))
    await db.commit()


@pytest.mark.asyncio
async def test_patch_zone(client):
    async with temporary_zone() as new_id:
        response = await client.patch(f'/zones/{new_id}', json={'name': 'farts'})
    assert response.status_code == 204


@pytest.mark.asyncio
async def test_delete_zone(client, new_zone):
    new_id = await new_zone()
    response = await client.delete(f'/zones/{new_id}')
    assert response.status_code == 204


@pytest.mark.asyncio
async def test_bad_post_patch_zone(client):
    response = await client.post("/zones", json={})
    assert response.status_code == 422
    async with temporary_zone() as new_id:
        response = await client.patch(f"/zones/{new_id}", json={})
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_get_zone(client):
    async with temporary_zone() as new_id:
        response = await client.get(f'/zones/{new_id}')
    assert response.status_code == 200


@pytest.mark.asyncio
async def test_get_zones(client):
    async with temporary_zone() as new_id:
        response = await client.get(f'/zones')
    assert response.status_code == 200
    zones = response.json()
    assert len(zones) == 1
    assert zones[0]['id'] == new_id


@pytest.mark.asyncio
async def test_zones_404(client):
    response = await client.get('/zones/1000')
    assert response.status_code == 404
    response = await client.patch('/zones/1000', json={'name': 'farts'})
    assert response.status_code == 404
    response = await client.delete('/zones/1000')
    assert response.status_code == 404



"""Modify history for start/end

Revision ID: 44f605b1cfa0
Revises: 92eaed4a95e3
Create Date: 2021-04-16 13:25:35.679663

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = '44f605b1cfa0'
down_revision = '92eaed4a95e3'
branch_labels = None
depends_on = None


def upgrade():
    op.execute("TRUNCATE TABLE history")
    op.execute("ALTER TABLE history RENAME ts TO end_ts")
    op.execute(
        "ALTER TABLE history ADD COLUMN start_ts timestamp without time zone "
        "DEFAULT (now() at time zone 'utc')")
    op.execute("ALTER TABLE history ALTER COLUMN end_ts SET DEFAULT NULL")
    op.execute("DROP FUNCTION append_history CASCADE")
    op.execute("""CREATE FUNCTION append_history()
        RETURNS TRIGGER
        LANGUAGE plpgsql
    AS $$
    BEGIN
        UPDATE history SET end_ts = (now() at time zone 'utc') WHERE tag_id = OLD.id AND end_ts IS NULL;
        INSERT INTO history (tag_id, zone_id, distance, zone_name)
        VALUES (NEW.id, NEW.zone_id, NEW.distance,
                (SELECT zones.name from zones where zones.id = NEW.zone_id));
        RETURN NEW;
    END;
    $$""")  # noqa: E501
    op.execute("""CREATE TRIGGER tags_update_zone
        AFTER UPDATE OF zone_id ON tags FOR EACH ROW
        WHEN (OLD.zone_id IS DISTINCT FROM NEW.zone_id)
        EXECUTE PROCEDURE append_history();""")
    op.execute("DELETE FROM history where start_ts IS NULL")


def downgrade():
    op.execute("DROP FUNCTION append_history CASCADE")
    op.execute("""CREATE FUNCTION append_history()
    RETURNS TRIGGER
    LANGUAGE plpgsql
AS $$
BEGIN
    INSERT INTO history (tag_id, zone_id, ts, distance, zone_name)
    VALUES (OLD.id, OLD.zone_id, OLD.last_seen, OLD.distance,
            (SELECT zones.name from zones where zones.id = OLD.zone_id));
    RETURN NEW;
END;
$$""")
    op.execute("""CREATE TRIGGER tags_update_zone
        AFTER UPDATE OF zone_id ON tags FOR EACH ROW
        WHEN (OLD.zone_id IS DISTINCT FROM NEW.zone_id AND OLD.zone_id IS NOT NULL)
        EXECUTE PROCEDURE append_history();""")  # noqa: E501
    op.execute(
        "ALTER TABLE history ALTER COLUMN end_ts "
        "SET DEFAULT (now() at time zone 'utc')")
    op.execute("ALTER TABLE history DROP COLUMN start_ts")
    op.execute("ALTER TABLE history RENAME end_ts TO ts")



from enum import IntEnum
from typing import Dict, Union

from pydantic import ConstrainedInt

Attributes = Dict[str, Union[str, int, bytes]]


class DbInteger(ConstrainedInt):
    min = 0
    max = 2**31 - 1


class Unsigned32Bit(ConstrainedInt):
    ge = 0
    lt = 2**32


class Unsigned16Bit(Unsigned32Bit):
    lt = 2**16


SensorValueType = Union[float, int]


class SensorType(IntEnum):
    SensorDataTemp = 0x01
    SensorDataHumid = 0x02
    SensorDataWetness = 0x4
    SensorDataAuth = 0xff


__all__ = ['Attributes', 'DbInteger', 'Unsigned16Bit', 'Unsigned32Bit',
           'SensorType', 'SensorValueType']



from typing import Iterable

from fastapi import APIRouter, Depends, Response  # type: ignore
import jinja2
import sqlalchemy as sa  # type: ignore
from sqlalchemy.ext.asyncio import AsyncConnection  # type: ignore

from ..db_next import get_db, tags, zones
import c3loc.api2.queries as q
from .dependencies import make_paginator, tag_filters
from .schemas.proximity import Proximity

router = APIRouter()


@router.get('/proximity', response_model=Iterable[Proximity])
async def proximity(db: AsyncConnection = Depends(get_db),
                    p=Depends(make_paginator),
                    filt=Depends(tag_filters)):
    result = await db.execute(p(filt(q.proximity())))
    return result

XML_TEMPLATE = jinja2.Template("""<?xml version="1.0" encoding="UTF-8"?>
<TagData>
  {%- for tag in tags %}
  <Tag id="{{ tag.id }}">
    <Name>{{ tag.name }}</Name>
    {%- if tag.zone_id %}
    <Location>{{ tag.zone_name }}</Location>
    <LocationId>{{ tag.zone_id }}</LocationId>
    <LastSeen>{{ tag.last_seen }}</LastSeen>
    <Distance>{{ tag.distance }}</Distance>
    {%- else %}
    <Location>Unknown</Location>
    {%- endif %}
  </Tag>
  {%- endfor %}
</TagData>
""")


@router.get('/proximity.xml')
async def proximity_xml(db: AsyncConnection = Depends(get_db)):
    query = sa.select(
        [tags.c.distance, zones.c.name.label('zone_name'),
         tags.c.name, tags.c.last_seen,
         tags.c.id, tags.c.zone_id, tags.c.alarm_active]
    ).select_from(tags.outerjoin(zones)).where(
        tags.c.type.in_(['SmartRelay', 'SecureSmartRelay'])
    ).order_by(tags.c.id.asc())
    results = (dict(r) for r in await db.execute(query))
    return Response(
        content=XML_TEMPLATE.render(tags=[dict(r) for r in results]),
        status_code=200, headers={'Content-Type': 'text/xml'})



import asyncio
import attr
from datetime import datetime
from typing import Callable, Dict, Optional

__all__ = ['run', 'register_cb', 'stop', 'increment', 'replace']


@attr.s
class Stat:
    count: int = 1
    pps_is_sensical: bool = True


stats: Dict[str, Stat] = {}
origin: datetime = datetime.now()
cbs: Dict[str, Callable[[], int]] = {}
stopped: Optional[asyncio.Future] = None
stop_requested = False
running = False
future: Optional[asyncio.Future] = None


def get_origin() -> datetime:
    global origin
    if not origin:
        origin = datetime.now()
    return origin


def run(interval):
    global future
    future = asyncio.ensure_future(main(interval))


async def main(interval) -> None:
    global running
    if running:
        return
    running = True
    get_origin()

    while True:
        await asyncio.sleep(interval)
        if stop_requested:
            break
        print_stats()
    running = False


def register_cb(name: str, cb: Callable[[], int]):
    cbs[name] = cb


async def stop():
    global stop_requested
    stop_requested = True
    if future:
        await future
    stop_requested = False


def print_stats() -> None:
    delta = (datetime.now() - get_origin())
    delta_s = delta.seconds
    out = f'*** Statistics @ {datetime.now()} ***\n'
    out += f'\tUptime: {delta}\n'
    for k, v in stats.items():
        if v.pps_is_sensical:
            out += f'\t{k}: {v.count} ({round(v.count/delta_s, 1)}/sec.)\n'
        else:
            out += f'\t{k}: {v.count}\n'
    for k, cb in cbs.items():
        out += f'\t{k}: {cb()}\n'
    out += '*** End Statistics ***'
    print(out)


def increment(k) -> None:
    if k in stats:
        stats[k].count += 1
    else:
        stats[k] = Stat()


def decrement(k) -> None:
    if k in stats:
        stats[k].count -= 0 if stats[k].count > 0 else 0
        stats[k].pps_is_sensical = False


def replace(k, v) -> None:
    if k not in stats:
        stats[k] = Stat()
    stats[k].count = v
    stats[k].pps_is_sensical = False



import click
from fastapi import FastAPI  # type: ignore
from fastapi.middleware.cors import CORSMiddleware  # type: ignore

import c3loc.api2.alarms as alarms
import c3loc.api2.groups as groups
import c3loc.api2.listeners as listeners
import c3loc.api2.proximity as proximity
import c3loc.api2.sensors as sensors
import c3loc.api2.tags as tags
import c3loc.api2.zones as zones

app = FastAPI()
app.include_router(alarms.router)
app.include_router(groups.router)
app.include_router(listeners.router)
app.include_router(proximity.router)
app.include_router(sensors.router)
app.include_router(tags.router)
app.include_router(zones.router)

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"]
)


@click.command()
@click.option('--port', '-p', default=8000)
def main(port: int) -> None:
    import uvicorn  # type: ignore  # nocov
    uvicorn.run(app, host="0.0.0.0", port=port)  # nosec  # nocov


if __name__ == '__main__':  # nocov
    main()



#!/bin/env python

import pkg_resources
from aiohttp import web
import aiohttp_jinja2
import jinja2


app = web.Application()
aiohttp_jinja2.setup(app, loader=jinja2.PackageLoader('c3loc', 'templates'))


@aiohttp_jinja2.template('index.html')
async def prox(request):
    return {}


@aiohttp_jinja2.template('tag_form.html')
async def tag_get(request):
    return {'tag_id': request.match_info['id']}


@aiohttp_jinja2.template('zone_form.html')
async def zone_get(request):
    return {'zone_id': request.match_info['id']}


@aiohttp_jinja2.template('add_zone.html')
async def add_zone(request):
    pass


app.router.add_static('/static/',
                      path=pkg_resources.resource_filename('c3loc', 'static'),
                      name='static')
app.router.add_get('/', prox)
app.router.add_get(r'/tags/{id:\d+}', tag_get)
app.router.add_get(r'/zones/{id:\d+}', zone_get)
app.router.add_get(r'/zones', add_zone)



"""log_relation_unlogged

Revision ID: 92eaed4a95e3
Revises: 720d3e9a9af0
Create Date: 2021-03-18 09:41:27.914377

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = '92eaed4a95e3'
down_revision = '720d3e9a9af0'
branch_labels = None
depends_on = None


def upgrade():
    op.execute("DROP TABLE log;")
    op.execute("""CREATE UNLOGGED TABLE log (
    tag_id int,
    zone_id int,
    ts timestamp without time zone default (now() at time zone 'utc'),
    distance numeric(4,2),
    variance real,
    listener_id varchar,
    data jsonb,
    anchor_dist numeric(4,2),
    anchor_ts_delta int,
    anchor_id int,
    reason report_reason,
    CONSTRAINT fk_tag
                 FOREIGN KEY (tag_id)
                 REFERENCES tags(id)
                 ON DELETE CASCADE,
    CONSTRAINT fk_zone
                 FOREIGN KEY (zone_id)
                 REFERENCES zones(id)
                 ON DELETE SET NULL,
    CONSTRAINT fk_anchor
                 FOREIGN KEY (anchor_id)
                 REFERENCES tags(id)
                 ON DELETE SET NULL,
    CONSTRAINT fk_listener
                 FOREIGN KEY (listener_id)
                 REFERENCES listeners(id)
                 ON DELETE SET NULL
);""")


def downgrade():
    pass



import attr
import typing
from enum import IntEnum


class SensorType(IntEnum):
    BatteryPercent = 0
    Temperature = 1
    Humidity = 2
    Alarm = 3
    TankHeight = 4
    TankConfidence = 5
    TemperaturePlus40 = 6
    AuthenticationTag = 0xff


@attr.s(frozen=True)
class SensorField:
    tag: int = attr.ib()
    length: int = attr.ib()
    value: bytes = attr.ib()

    @classmethod
    def from_bytes(cls, data):
        if len(data) < 2:
            return None
        tag = data[0]
        length = data[1]
        if length > len(data):
            return None
        return cls(tag, length, data[2:2+length])

    def size(self):
        return self.length + 2


@attr.s(frozen=True)
class SensorReport:
    fields: typing.List[SensorField] = attr.ib(default=[])

    @classmethod
    def parse(cls, data):
        fields = []
        if len(data) < 2:
            sensor = cls()
            return sensor
        while data:
            field = SensorField.from_bytes(data)
            if not field:
                break
            fields.append(field)
            data = data[field.size():]
        return cls(fields)



"""Adds gist tsrange index to history

Revision ID: 515df7865792
Revises: dbe74359375f
Create Date: 2021-07-10 14:25:16.407379

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = '515df7865792'
down_revision = 'dbe74359375f'
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        'create index tsrange_idx on history '
        'using gist (tsrange(start_ts, end_ts))')


def downgrade():
    op.execute('drop index history.tsrange_idx')



import re
from datetime import datetime, timedelta
from enum import Enum
from typing import Optional, Dict, Type, Union
from uuid import UUID

from pydantic import (  # type: ignore
    BaseModel, ConstrainedStr, Field, root_validator,
    validator)

from ...config import CONFIG
from .types import DbInteger, Unsigned16Bit, Unsigned32Bit


class TagType(str, Enum):
    SmartRelay = 'SmartRelay'
    SecureSmartRelay = 'SecureSmartRelay'
    LocationAnchor = 'LocationAnchor'
    IBeacon = 'iBeacon'
    AlertWet = 'AlertWet'
    MacBeacon = 'MacBeacon'


class TagBase(BaseModel):
    type: TagType
    name: Optional[str]
    zone_id: Optional[DbInteger]
    group_id: Optional[DbInteger]
    attrs: Optional[Dict[str, Union[str, int]]] = {}


class MacAddress(ConstrainedStr):
    regex = re.compile(
        r'^[a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2}:'
        r'[a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2}$')


class TagCreate(TagBase):
    uuid: Optional[UUID]
    major: Optional[Unsigned16Bit]
    minor: Optional[Unsigned16Bit]
    mac: Optional[MacAddress]
    bid: Optional[Unsigned32Bit]


class Tag(TagBase):
    last_seen: datetime
    zone_start: datetime
    battery_pct: Optional[int]
    alarm_active: Optional[bool]


class IBeacon(TagBase):
    uuid: UUID
    major: Unsigned16Bit
    minor: Unsigned16Bit
    type: TagType = Field(default=TagType.IBeacon, const=True)

    @root_validator
    def default_name(cls, values: Dict) -> Dict:
        if values['name']:
            return values
        uuid = values['uuid']
        major = values['major']
        minor = values['minor']
        if uuid == CONFIG['LA_UUID']:
            values['name'] = f'LocationAnchor {uuid}:{major}:{minor}'
        else:
            values['name'] = f'iBeacon {uuid}:{major}:{minor}'
        return values


class AlertWet(TagBase):
    bid: Unsigned32Bit
    type: TagType = Field(default=TagType.AlertWet, const=True)

    @root_validator
    def default_name(cls, values: Dict) -> Dict:
        if values['name']:
            return values
        values['name'] = f'AlertWet {hex(values["bid"])}'
        return values


class MacBeacon(TagBase):
    bid: Unsigned32Bit
    type: TagType = Field(default=TagType.MacBeacon, const=True)

    @root_validator
    def default_name(cls, values: Dict) -> Dict:
        if values['name']:
            return values
        values['name'] = f'MacBeacon {hex(values["bid"])}'
        return values


class SmartRelay(TagBase):
    mac: MacAddress = Field(...,)
    type: TagType = Field(default=TagType.SmartRelay, const=True)

    @root_validator
    def default_name(cls, values: Dict) -> Dict:
        if values['name']:
            return values
        values['name'] = f'SmartRelay {values["mac"]}'
        return values


class SecureSmartRelay(TagBase):
    bid: Unsigned32Bit
    type: TagType = Field(default=TagType.SecureSmartRelay, const=True)

    @root_validator
    def default_name(cls, values: Dict) -> Dict:
        if values['name']:
            return values
        values['name'] = f'SecureSmartRelay {hex(values["bid"])}'
        return values


class TagPatch(BaseModel):
    name: Optional[str]
    zone_id: Optional[DbInteger]
    group_id: Optional[DbInteger]
    attrs: Optional[Dict[str, Union[str, int]]] = {}


def dispatch_model(create: TagCreate):
    map: Dict[TagType, Type[BaseModel]] = {
        TagType.IBeacon: IBeacon,
        TagType.SecureSmartRelay: SecureSmartRelay,
        TagType.SmartRelay: SmartRelay,
        TagType.AlertWet: AlertWet,
        TagType.MacBeacon: MacBeacon}

    return map[create.type].parse_obj(create.dict())


class TagHistory(BaseModel):
    zone_id: DbInteger
    start_ts: datetime
    end_ts: Optional[datetime] = None
    distance: float
    duration: Optional[timedelta] = None
    zone_name: Optional[str]

    @validator('end_ts', always=True)
    def default_end_ts(cls, v: datetime) -> datetime:
        if v:
            return v
        return datetime.utcnow()

    @validator('duration', always=True)
    def calc_duration(cls, _, values):
        return values['end_ts'] - values['start_ts']


class TagContact(BaseModel):
    tag_id: DbInteger
    tag_name: str
    end_ts: datetime
    start_ts: datetime
    duration: Optional[timedelta] = None
    zone_id: DbInteger
    zone_name: str
    hash: str

    @validator('duration', always=True)
    def calc_duration(cls, _, values):
        return values['end_ts'] - values['start_ts']



from __future__ import annotations

from sqlalchemy.ext.asyncio import AsyncConnection  # type: ignore
from typing import List

from fastapi import (  # type: ignore
    APIRouter, Depends, HTTPException, Request, Response)
from pydantic import ValidationError  # type: ignore

from ..db_next import get_db, tags, history
from .dependencies import (make_paginator, tag_filters, time_filters,
                           make_contacts_paginator, bindable_time_filter)
import c3loc.api2.queries as q
from .schemas.tags import (Tag, TagContact, TagCreate, TagHistory, TagPatch,
                           dispatch_model)


router = APIRouter()


@router.post('/tags')
async def create_tag(new_tag: TagCreate, request: Request,
                     db: AsyncConnection = Depends(get_db)):
    try:
        new_tag = dispatch_model(new_tag)
    except ValidationError as e:
        raise HTTPException(status_code=422, detail=e.json())
    insert = tags.insert().values(new_tag.dict())
    result = await db.execute(insert)
    new_id = result.inserted_primary_key[0]
    root_path = request.scope.get("root_path")
    return Response(status_code=201, headers={
        'Location': f'{root_path}/tags/{new_id}'})


@router.get('/tags', response_model=List[Tag])
async def get_tags(db: AsyncConnection = Depends(get_db),
                   p=Depends(make_paginator),
                   filt=Depends(tag_filters)):
    result = await db.execute(p(filt(q.get_tags())))
    return result


@router.delete('/tags/{tag_id}')
async def delete_tag(tag_id: int, db: AsyncConnection = Depends(get_db)):
    result = await db.execute(tags.delete().where(tags.c.id == tag_id))
    if result.rowcount == 0:
        raise HTTPException(404)
    return Response(status_code=204)


@router.get('/tags/{tag_id}', response_model=Tag)
async def get_tag(tag_id: int, db: AsyncConnection = Depends(get_db)):
    entry = await q.get_or_404(db, tags, tag_id)
    return entry


@router.patch('/tags/{tag_id}')
async def patch_tag(tag_id: int, tag_updates: TagPatch,
                    db: AsyncConnection = Depends(get_db)):
    await q.get_or_404(db, tags, tag_id)
    await db.execute(
        tags.update().values(tag_updates.dict()).where(tags.c.id == tag_id))
    return Response(status_code=204)


@router.get('/tags/{tag_id}/history', response_model=List[TagHistory])
async def tag_history(tag_id: int,
                      db: AsyncConnection = Depends(get_db),
                      p=Depends(make_paginator),
                      t=Depends(time_filters)):
    await q.get_or_404(db, tags, tag_id)
    query = q.get_history(tag_id)
    query = t(query, history.c.start_ts, history.c.end_ts)
    return (TagHistory.parse_obj(r) for r in await db.execute(p(query)))


@router.get('/tags/{tag_id}/contacts', response_model=List[TagContact])
async def tag_contacts(tag_id: int,
                       db: AsyncConnection = Depends(get_db),
                       p=Depends(make_contacts_paginator),
                       t=Depends(bindable_time_filter)):
    await q.get_or_404(db, tags, tag_id)
    query = q.get_contacts(tag_id)
    print(query)
    query = t(query)
    print(query)
    query = p(query)
    print(query)
    results = await db.execute(query)
    return (TagContact.parse_obj(r) for r in results)



from datetime import datetime
from typing import Optional

from pydantic import BaseModel  # type: ignore


class AlarmLinks(BaseModel):
    tag: str
    zone: str


class Alarm(BaseModel):
    id: int
    tag_name: str
    zone_name: Optional[str]
    start_ts: datetime
    last_ts: datetime
    acknowledged: bool
    priority: int


class AlarmPatch(BaseModel):
    acknowledged: bool



import sqlalchemy as sa  # type: ignore
from sqlalchemy.ext.asyncio import AsyncConnection  # type: ignore
from fastapi import HTTPException

from ..db_next import alarms, history, sensors, tags, zones


def get_alarm(alarm_id: int) -> sa.sql.expression:
    return get_alarms().where(alarms.c.id == alarm_id)


def get_alarms() -> sa.sql.expression:
    return sa.select([alarms.c.id, tags.c.name.label('tag_name'),
                      zones.c.name.label('zone_name'),
                      alarms.c.start_ts, alarms.c.last_ts,
                      alarms.c.acknowledged, alarms.c.priority]).select_from(
        alarms.join(tags).outerjoin(zones))


def get_tags() -> sa.sql.expression:
    return sa.select(tags)


def proximity() -> sa.sql.expression:
    return sa.select(
        [tags.c.distance, zones.c.name.label('zone_name'),
         tags.c.name.label('tag_name'), tags.c.type.label('tag_type'),
         tags.c.last_seen, tags.c.id.label('tag_id'), tags.c.zone_id,
         tags.c.alarm_active]).select_from(
        tags.outerjoin(zones)).order_by(tags.c.id.desc())


def get_sensors() -> sa.sql.expression:
    return sa.select(
        [sensors.c.tag_id, sensors.c.type, sensors.c.id,
         sensors.c.ts, sensors.c.value]).order_by(
        sensors.c.tag_id, sensors.c.type, sensors.c.ts.desc()).distinct(
        sensors.c.tag_id, sensors.c.type)


def get_history(tag_id: int):
    return sa.select([history.c.tag_id, history.c.zone_id, history.c.start_ts,
                      history.c.distance, history.c.end_ts,
                      zones.c.name.label('zone_name')]).select_from(
        history.outerjoin(zones)).order_by(
            history.c.start_ts.desc()).where(history.c.tag_id == tag_id)


# def get_contacts(tag_id: int):
#     other_tag_hist = sa.select([
#         history.c.tag_id,
#         history.c.zone_id,
#         history.c.zone_name,
#         history.c.start_ts,
#         tags.c.name.label('tag_name'),
#         sa.func.coalesce(
#             history.c.end_ts,
#             sa.text('(now() at time zone \'utc\')')
#         ).label('end_ts')
#     ]).join(tags, history.c.tag_id == tags.c.id).where(sa.sql.and_(
#         history.c.tag_id != tag_id,
#         tags.c.type.in_(['SmartRelay', 'SecureSmartRelay']))).cte('other')
#     return sa.select([
#         other_tag_hist.c.tag_id,
#         other_tag_hist.c.zone_id,
#         other_tag_hist.c.zone_name.label('zone_name'),
#         other_tag_hist.c.tag_name,
#         sa.func.greatest(
#             history.c.start_ts,
#             other_tag_hist.c.start_ts).label('start_ts'),
#         sa.func.least(
#              history.c.end_ts, other_tag_hist.c.end_ts
#         ).label('end_ts'),
#         sa.func.md5(
#             sa.func.concat(
#                 tag_id,
#                 other_tag_hist.c.tag_id,
#                 other_tag_hist.c.zone_id,
#                 sa.func.greatest(
#                     history.c.start_ts,
#                     other_tag_hist.c.start_ts).label('start_ts'),
#                 sa.func.least(
#                     history.c.end_ts, other_tag_hist.c.end_ts))).label(
#             'hash')
#     ]).select_from(other_tag_hist).join(
#         history, sa.tuple_(history.c.start_ts, history.c.end_ts).op(
#             'overlaps')(sa.tuple_(other_tag_hist.c.start_ts,
#                                   other_tag_hist.c.end_ts))).where(
#         history.c.tag_id == tag_id)

def get_contacts(tag_id: int):
    q = sa.text("""
with target_tag as (
    SELECT tag_id, start_ts, end_ts, zone_id
    from history
    where tag_id = :tag_id
      and tsrange(start_ts, end_ts) && tsrange(:start, :end)
),
     other_tag as (
         SELECT tag_id, start_ts, end_ts, zone_id
         from history
         where tag_id != :tag_id
           and tsrange(start_ts, end_ts) && tsrange(:start, :end)
     ),
     overlap as (
         SELECT other_tag.tag_id,
                greatest(target_tag.start_ts, other_tag.start_ts)                                as start_ts,
                least(coalesce(target_tag.end_ts, (now() at time zone 'utc')), other_tag.end_ts) as end_ts,
                other_tag.zone_id                                                                as zone_id,
                least(coalesce(target_tag.end_ts, (now() at time zone 'utc')), other_tag.end_ts) -
                greatest(target_tag.start_ts, other_tag.start_ts)                                as duration
         from target_tag,
              other_tag
         where target_tag.zone_id = other_tag.zone_id
           and tsrange(target_tag.start_ts, target_tag.end_ts) && tsrange(other_tag.start_ts, other_tag.end_ts)
     )
SELECT overlap.*,
       zones.name as zone_name,
       tags.name as tag_name,
       md5(concat(cast(:tag_id_str as text), tag_id::text, overlap.zone_id,
       start_ts::text,
       end_ts::text)) as hash
from overlap
         join tags on tag_id = tags.id
         join zones on zones.id = overlap.zone_id
         order by start_ts desc limit :limit offset :offset"""  # noqa: E501
                )
    return q.bindparams(tag_id=tag_id, tag_id_str=str(tag_id))


async def get_or_404(db: AsyncConnection,
                     table: sa.Table, t_id: int):
    result = await db.execute(table.select().where(table.c.id == t_id))
    entry = result.first()
    if not entry:
        raise HTTPException(404)
    return entry



from typing import Iterable

import fastapi  # type: ignore
from fastapi import Depends, APIRouter, Request
import sqlalchemy as sa  # type: ignore
from sqlalchemy.ext.asyncio import AsyncConnection  # type: ignore

from ..db_next import zones, get_db
from .schemas.zones import Zone, ZoneCreate, ZonePatch

router = APIRouter()


@router.get('/zones', response_model=Iterable[Zone])
async def get_zones(db: AsyncConnection = Depends(get_db)):
    return await db.execute(sa.select(zones))


@router.post('/zones')
async def create_zone(zone: ZoneCreate, request: Request,
                      db: AsyncConnection = Depends(get_db)):
    insert = zones.insert().values(zone.dict())
    result = await db.execute(insert)
    new_id = result.inserted_primary_key[0]
    root_path = request.scope.get("root_path")
    return fastapi.Response(status_code=201, headers={
        'Location': f'{root_path}/zones/{new_id}'})


@router.get('/zones/{zone_id}', response_model=Zone)
async def get_zone(zone_id: int, db: AsyncConnection = Depends(get_db)):
    result = await db.execute(zones.select().where(zones.c.id == zone_id))
    entry = result.first()
    if not entry:
        raise fastapi.HTTPException(404)
    return entry


@router.delete('/zones/{zone_id}')
async def delete_zone(zone_id: int, db: AsyncConnection = Depends(get_db)):
    result = await db.execute(zones.delete().where(zones.c.id == zone_id))
    if result.rowcount == 0:
        raise fastapi.HTTPException(404)
    return fastapi.Response(status_code=204)


@router.patch('/zones/{zone_id}')
async def patch_zone(zone_id: int, zone_updates: ZonePatch,
                     db: AsyncConnection = Depends(get_db)):
    result = await db.execute(zones.update().values(
        zone_updates.dict()).where(zones.c.id == zone_id))
    if result.rowcount == 0:
        raise fastapi.HTTPException(404)
    return fastapi.Response(status_code=204)



import contextlib
from typing import Tuple

import pytest
from httpx import AsyncClient
from hypothesis import given, settings, strategies as st

from c3loc.api2.main import app
from c3loc.api2.schemas.types import SensorType
from c3loc.db_next import sensors

from .conftest import get_engine
from .test_api2_tags import temporary_sr_tag


@pytest.fixture
async def add_sensor(engine):
    add_sensor.ids = set()
    async def f(t_id: int, t: SensorType, v: bytes) -> None:
        async with engine.connect() as conn:
            insert = sensors.insert().values({
                'tag_id': t_id,
                'type': t,
                'value': v,
                'device_ts': 0})
            r = await conn.execute(insert)
            add_sensor.ids.add(r.inserted_primary_key[0])
            await conn.commit()
    yield f


@pytest.mark.asyncio
async def test_dont_barf_on_unk_sensor_type(add_sensor, client, engine):
    async with temporary_sr_tag() as tag_id:
        await add_sensor(tag_id, 254, b'\x00')
        try:
            response = await client.get('/sensors')
            assert response.status_code == 200
            assert len(response.json()) == 0
        finally:
            async with engine.connect() as conn:
                await conn.execute(sensors.delete())
                await conn.commit()


@pytest.mark.asyncio
async def test_only_latest_sensor(add_sensor, client, engine):
    async with temporary_sr_tag() as tag_id:
        await add_sensor(tag_id, SensorType.SensorDataHumid, b'\x00')
        await add_sensor(tag_id, SensorType.SensorDataHumid, b'\x01')
        try:
            response = await client.get('/sensors')
            assert response.status_code == 200
            assert len(response.json()) == 1
            assert response.json()[0]['value'] == 1
        finally:
            async with engine.connect() as conn:
                await conn.execute(sensors.delete())
                await conn.commit()


@contextlib.asynccontextmanager
async def mock_sensor(typ: int = SensorType.SensorDataTemp,
                      value: bytes = b'\x00'):
    async with get_engine().connect() as conn:
        async with temporary_sr_tag() as tag_id:
            insert = sensors.insert().values({
                'tag_id': tag_id,
                'type': typ,
                'value': value,
                'device_ts': 0})
            r = await conn.execute(insert)
            sensor_id = r.inserted_primary_key[0]
            await conn.commit()
            try:
                yield tag_id
            finally:
                await conn.execute(sensors.delete().where(
                    sensors.c.id == sensor_id))
                await conn.commit()


@pytest.mark.asyncio
async def test_get_sensors_null(client):
    response = await client.get("/sensors")
    assert response.status_code == 200
    assert response.json() == []


@pytest.mark.asyncio
async def test_get_sensors(client):
    async with mock_sensor() as tag_id:
        response = await client.get(f'/sensors')
        assert response.status_code == 200
        ls = response.json()
        assert len(ls) == 1
        assert ls[0]['tag_id'] == tag_id

        response = await client.get('/proximity?offset=1')
        assert response.status_code == 200
        ls = response.json()
        assert len(ls) == 0

        response = await client.get('/listeners?limit=0')
        assert response.status_code == 200
        ls = response.json()
        assert len(ls) == 0


def make_raw_temp(temp: int) -> Tuple[int, bytes]:
    if temp > 2**7-1 or temp < -2**7:
        return temp, temp.to_bytes(2, 'little', signed=True)
    return temp, temp.to_bytes(1, 'little', signed=True)


@given(temp=st.builds(
    make_raw_temp, st.integers(min_value=-2732, max_value=2**15-1)))
@settings(report_multiple_bugs=False, deadline=None)
@pytest.mark.asyncio
async def test_sensors_temp(temp: Tuple[int, bytes]):
    target, raw_val = temp
    target = round(target / 10.0, 1)
    async with mock_sensor(SensorType.SensorDataTemp, raw_val) as tag_id:
        async with AsyncClient(app=app, base_url='http://test') as client:
            response = await client.get('/sensors/')
        assert response.status_code == 200
        assert response.json()[0]['value'] == target


def make_raw_humid(humid: int) -> bytes:
    return humid.to_bytes(1, 'little', signed=False)


@given(humid=st.builds(
    make_raw_humid, st.integers(min_value=0, max_value=100)))
@settings(report_multiple_bugs=False, deadline=None)
@pytest.mark.asyncio
async def test_sensors_humid(humid: bytes):
    target = int.from_bytes(humid, 'little', signed=False)
    async with mock_sensor(SensorType.SensorDataHumid, humid) as tag_id:
        async with AsyncClient(app=app, base_url='http://test') as client:
            response = await client.get('/sensors/')
        assert response.status_code == 200
        assert response.json()[0]['value'] == target


# @pytest.mark.asyncio
# async def test_proximity_filter_type(client):
#     async with temporary_sr_tag() as new_id:
#         response = await client.get('/proximity?type=SmartRelay')
#         assert response.status_code == 200
#         assert response.json()[0]['tag_id'] == new_id
#
#         response = await client.get('/proximity?type=iBeacon')
#         print('Response', response.json())
#         assert response.status_code == 200
#         assert len(response.json()) == 0



from datetime import datetime, timedelta
import json

from aiohttp import web
from aiohttp_cors import CorsViewMixin  # type: ignore
import asyncpg  # type: ignore
from dateutil.parser import isoparse
from marshmallow import EXCLUDE
import pytz

from ..config import CONFIG
from .schemas import (ibeacon_post_schema, macbeacon_post_schema,
                      tag_patch_schema, history_schema, contacts_schema,
                      secure_beacon_post_schema)
from .sensors import format_sensors
from .views import ValidatingView, paginate_query


def tag_schema_dump(row):
    if row['type'] in {'iBeacon', 'LocationAnchor'}:
        return ibeacon_post_schema.dump(row)
    if row['type'] in {'SecureLocationAnchor', 'SecureSmartRelay'}:
        return secure_beacon_post_schema.dump(row)
    return macbeacon_post_schema.dump(row)


class TagsView(ValidatingView):
    async def get(self):
        async with self.request.app['db_pool'].acquire() as conn:
            query = None
            if 'type' in self.request.query:
                query = ('SELECT * from tags WHERE type = $1',
                         self.request.query['type'])
            elif 'group_id' in self.request.query:
                try:
                    query = ('SELECT * from tags WHERE group_id = $1',
                             int(self.request.query['group_id']))
                except ValueError:
                    raise web.HTTPBadRequest(text='bad group_id')
            else:
                query = ('SELECT * from tags', )
            query = paginate_query(self.request, query)
            try:
                tags = await conn.fetch(*query)
            except asyncpg.exceptions.InvalidTextRepresentationError:
                raise web.HTTPBadRequest(text='Invalid tag type')
            return web.json_response(
                [tag_schema_dump(r) for r in tags if r is not None])

    async def post(self):
        body = await self._valid_json(self.request)

        if 'type' not in body:
            raise web.HTTPBadRequest(reason='Type field is required for post')

        new_t = {}
        if body['type'] in {'iBeacon', 'LocationAnchor'}:
            new_t.update(self._validate(ibeacon_post_schema, body))
        elif body['type'] == 'SmartRelay':
            new_t.update(self._validate(macbeacon_post_schema, body))
        elif body['type'] in {'SecureSmartRelay', 'SecureLocationAnchor'}:
            new_t.update(self._validate(secure_beacon_post_schema, body))
        else:
            raise web.HTTPBadRequest(reason=f'Unknown tag type {body["type"]}')

        async with self.request.app['db_pool'].acquire() as conn:
            try:
                await conn.execute(
                    'INSERT INTO tags '
                    '(mac, uuid, major, minor, type, zone_id, name, '
                    'group_id, bid, attrs) VALUES '
                    '($1, $2, $3, $4, $5, $6, $7, $8, $9, $10)',
                    new_t.get('mac', None),
                    new_t.get('uuid', None),
                    new_t.get('major', None),
                    new_t.get('minor', None),
                    new_t['type'],
                    new_t.get('zone_id', None),
                    new_t.get('name', None),
                    new_t.get('group_id', None),
                    new_t.get('bid', None),
                    json.dumps(new_t.get('attrs', {})))
            except asyncpg.exceptions.UniqueViolationError as e:
                raise web.HTTPConflict(reason=str(e.detail))
            except (asyncpg.exceptions.InvalidTextRepresentationError,
                    asyncpg.exceptions.NotNullViolationError,
                    asyncpg.exceptions.ForeignKeyViolationError) as e:
                raise web.HTTPBadRequest(reason=str(e.detail))
        raise web.HTTPCreated


def parse_start_end(request):
    utc = pytz.timezone('UTC')
    if 'start' in request.query:
        try:
            start = isoparse(request.query['start'])
        except ValueError:
            raise web.HTTPBadRequest(
                reason="Invailid ISO8601 in start parameter")
        if start.tzinfo:
            start = start.astimezone(utc).replace(tzinfo=None)
    else:
        start = datetime.utcnow() - timedelta(days=7)

    if 'end' in request.query:
        try:
            end = isoparse(request.query['end'])
        except ValueError:
            raise web.HTTPBadRequest(
                reason="Invailid ISO8601 in end parameter")
        if end.tzinfo:
            end = end.astimezone(utc).replace(tzinfo=None)
    else:
        end = datetime.max

    return start, end


HISTORY_QUERY = """
SELECT tag_id,
       zone_id,
       start_ts,
       distance,
       coalesce(end_ts, (now() at time zone 'utc'))            as end_ts,
       coalesce(end_ts, (now() at time zone 'utc')) - start_ts as duration
FROM history
WHERE tag_id = $1 AND tsrange(start_ts, end_ts) && tsrange($2, $3)
ORDER BY start_ts DESC
"""  # noqa: E501


class TagHistoryView(web.View, CorsViewMixin):
    async def get(self):
        t_id = int(self.request.match_info['id'])
        start, end = parse_start_end(self.request)
        async with self.request.app['db_pool'].acquire() as conn:
            query = (HISTORY_QUERY, t_id, start, end)
            query = paginate_query(self.request, query)
            history = await conn.fetch(*query)
            return web.json_response(history_schema.dump(history))


CONTACT_QUERY = """
with target_tag as (
    SELECT tag_id, start_ts, end_ts, zone_id
    from history
    where tag_id = $1
      and tsrange(start_ts, end_ts) && tsrange($2, $3)
),
     other_tag as (
         SELECT tag_id, start_ts, end_ts, zone_id
         from history
         where tag_id != $1
           and tsrange(start_ts, end_ts) && tsrange($2, $3)
     ),
     overlap as (
         SELECT other_tag.tag_id,
                greatest(target_tag.start_ts, other_tag.start_ts)                                as start_ts,
                least(coalesce(target_tag.end_ts, (now() at time zone 'utc')), other_tag.end_ts) as end_ts,
                other_tag.zone_id                                                                as zone_id,
                least(coalesce(target_tag.end_ts, (now() at time zone 'utc')), other_tag.end_ts) -
                greatest(target_tag.start_ts, other_tag.start_ts)                                as duration
         from target_tag,
              other_tag
         where target_tag.zone_id = other_tag.zone_id
           and tsrange(target_tag.start_ts, target_tag.end_ts) && tsrange(other_tag.start_ts, other_tag.end_ts)
     )
SELECT overlap.*,
       zones.name as zone_name,
       tags.name as tag_name,
       md5(concat($1::text, tag_id::text, overlap.zone_id, start_ts::text, end_ts::text)) as hash
from overlap
         join tags on tag_id = tags.id
         join zones on zones.id = overlap.zone_id
         order by start_ts desc
"""  # noqa: E501


async def tag_exists_or_404(t_id: int, conn) -> None:
    tag_exists = await conn.fetchval(
        "SELECT id from tags where id = $1", t_id)
    if not tag_exists:
        raise web.HTTPNotFound


class TagContactView(web.View, CorsViewMixin):
    async def get(self):
        t_id = int(self.request.match_info['id'])
        start, end = parse_start_end(self.request)
        async with self.request.app['db_pool'].acquire() as conn:
            await tag_exists_or_404(t_id, conn)
            query = (CONTACT_QUERY, t_id, start, end)
            max_limit = int(CONFIG['API_RESULT_LIMIT_CONTACTS'])
            query = paginate_query(self.request, query, max_limit=max_limit)
            contacts = await conn.fetch(*query)
            return web.json_response(contacts_schema.dump(contacts))


SENSORS_QUERY = """
SELECT * FROM sensors
WHERE tag_id = $1 AND ts >= $2 AND ts <= $3 ORDER BY ts DESC
"""


class TagSensorView(web.View, CorsViewMixin):
    async def get(self):
        t_id = int(self.request.match_info['id'])
        start, end = parse_start_end(self.request)
        async with self.request.app['db_pool'].acquire() as conn:
            await tag_exists_or_404(t_id, conn)
            query = (SENSORS_QUERY, t_id, start, end)
            query = paginate_query(self.request, query)
            sensors = await conn.fetch(*query)
            return web.json_response(format_sensors(sensors))


class TagView(ValidatingView):
    async def get(self):
        t_id = int(self.request.match_info['id'])
        async with self.request.app['db_pool'].acquire() as conn:
            tag = await conn.fetchrow(
                'SELECT * FROM tags t WHERE id = $1', t_id)
            if not tag:
                raise web.HTTPNotFound
            return web.json_response(tag_schema_dump(tag))

    async def patch(self):
        t_id = int(self.request.match_info['id'])
        async with self.request.app['db_pool'].acquire() as conn:
            row = await conn.fetchrow(
                'SELECT * FROM tags t WHERE id = $1', t_id)
            if not row:
                raise web.HTTPNotFound
            tag = dict(row.items())
            body = await self._valid_json(self.request)
            updates = self._validate(tag_patch_schema, body)
            if 'type' in updates:
                # Ensure that we are validating against existing type
                del body['type']
            tag.update(updates)
            if tag['type'] in {'iBeacon', 'LocationAnchor'}:
                self._validate(ibeacon_post_schema, tag, unknown=EXCLUDE)
            elif tag['type'] == 'SmartRelay':
                self._validate(macbeacon_post_schema, tag, unknown=EXCLUDE)
            elif tag['type'] in {'SecureSmartRelay', 'SecureLocationAnchor'}:
                self._validate(secure_beacon_post_schema, tag, unknown=EXCLUDE)
            else:
                raise web.HTTPBadRequest(
                    reason=f'Unknown tag type {body["type"]}')
            await conn.execute(
                'UPDATE tags SET name = $1, mac = $2, uuid = $3, major = $4, '
                'minor = $5, attrs = $6, zone_id = $7, group_id = $8, '
                'bid = $9 WHERE id = $10', tag['name'], tag['mac'],
                tag['uuid'], tag['major'], tag['minor'],
                json.dumps(tag['attrs']), tag['zone_id'], tag['group_id'],
                tag['bid'], t_id)
        raise web.HTTPNoContent

    async def delete(self):
        t_id = int(self.request.match_info['id'])
        async with self.request.app['db_pool'].acquire() as conn:
            await conn.execute('DELETE from tags where id = $1', t_id)
        raise web.HTTPNoContent
